#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np
from scipy import linalg
from pyscf.lib import logger


def _reshape_vxc_sigma(vxc0, dens_deriv):
    # d/drho, d/dsigma -> d/drho, d/drho'
    vrho = vxc0[0]
    vxc1 = list(vrho.T)
    if dens_deriv:
        vsigma = vxc0[1]
        vxc1 = vxc1 + list(vsigma.T)
    else:
        vxc1 = [vxc1[0][None, :], vxc1[1][None, :]]
    return vxc1


def _unpack_vxc_sigma(vxc0, rho, dens_deriv):
    # d/drho, d/dsigma -> d/drho, d/drho'
    vrho = vxc0[0]
    vxc1 = list(vrho.T)
    if dens_deriv:
        vsigma = vxc0[1]
        vxc1 = vxc1 + list(vsigma.T)
        vxc1 = _unpack_sigma_vector(vxc1, rho[0][1:4], rho[1][1:4])
    else:
        vxc1 = [vxc1[0][None, :], vxc1[1][None, :]]
    return vxc1


def _pack_fxc_ltri(fxc0, dens_deriv):
    # d2/drho2, d2/drhodsigma, d2/dsigma2
    # -> lower-triangular Hessian matrix
    frho = fxc0[0].T
    fxc1 = [frho[0], ]
    fxc1 += [frho[1], frho[2], ]
    if dens_deriv:
        frhosigma, fsigma = fxc0[1].T, fxc0[2].T
        fxc1 += [frhosigma[0], frhosigma[3], fsigma[0], ]
        fxc1 += [frhosigma[1], frhosigma[4], fsigma[1], fsigma[3], ]
        fxc1 += [frhosigma[2], frhosigma[5], fsigma[2], fsigma[4], fsigma[5]]
    return fxc1


def eval_ot(otfnal, rho, Pi, dderiv=1, weights=None, _unpack_vot=True):
    r'''get the integrand of the on-top xc energy and its functional
    derivatives wrt rho and Pi

    Args:
        rho : ndarray of shape (2,*,ngrids)
            containing spin-density [and derivatives]
        Pi : ndarray with shape (*,ngrids)
            containing on-top pair density [and derivatives]

    Kwargs:
        dderiv : integer
            Order of derivatives to return
        weights : ndarray of shape (ngrids)
            used ONLY for debugging the total number of ``translated''
            electrons in the calculation of rho_t
            Not multiplied into anything!
        _unpack_vot : logical
            If True, derivatives with respect to density gradients are
            reported as de/drho' and de/dPi'; otherwise, they are
            reported as de/d|rho'|^2, de/d(rho'.Pi'), and de/d|Pi'|^2

    Returns:
        eot : ndarray of shape (ngrids)
            integrand of the on-top exchange-correlation energy
        vot : ndarrays of shape (*,ngrids) or None
            first functional derivative of Eot wrt (density, pair
            density) and their derivatives. If _unpack_vot = True, shape
            and format is ([a, ngrids], [b, ngrids]) : (vrho, vPi);
            otherwise, [c, ngrids] : [rho,Pi,|rho'|^2,tau,rho'.Pi',|Pi'|^2]
            ftGGA: a=4, b=4, c=5 (drop tau)
            tmGGA: a=5, b=1, c=4 (drop Pi')
            tGGA: a=4, b=1, c=3 (drop Pi', tau)
            *tLDA: a=1, b=1, c=2 (drop rho', tau)
        fot : ndarray of shape (*,ngrids) or None
            second functional derivative of Eot wrt density, pair
            density, and derivatives; first dimension is lower-
            triangular matrix elements corresponding to the basis
            (rho, Pi, |rho'|^2, rho'.Pi', |Pi'|^2)
            stopping at Pi (3 elements) for *tLDA and |rho'|^2 (6
            elements) for tGGA.
    '''
    if dderiv > 2:
        raise NotImplementedError("Translation of density derivatives of "
                                  "higher order than 2")
    if rho.ndim == 2: rho = rho[:, None, :]
    if Pi.ndim == 1: Pi = Pi[None, :]
    assert (rho.shape[0] == 2)
    assert (rho.shape[1] <= 6), "Undefined behavior for this function"

    nderiv = rho.shape[1]
    nderiv_Pi = Pi.shape[0]

    rho_t = otfnal.get_rho_translated(Pi, rho)
    # LDA in libxc has a special numerical problem with zero-valued densities
    # in one spin
    if nderiv == 0:
        idx = (rho_t[0, 0] > 1e-15) & (rho_t[1, 0] < 1e-15)
        rho_t[1, 0, idx] = 1e-15
        idx = (rho_t[0, 0] < 1e-15) & (rho_t[1, 0] > 1e-15)
        rho_t[0, 0, idx] = 1e-15

    # mGGA in libxc has special numerical problem with zero-valued densities in
    # one spin!
    if nderiv == 5:
        idx = (rho_t[0, 4] > 1e-15) & (rho_t[1, 4] < 1e-15)
        rho_t[1, 4, idx] = 1e-15
        idx = (rho_t[0, 4] < 1e-15) & (rho_t[1, 4] > 1e-15)
        rho_t[0, 4, idx] = 1e-15

    rho_tot = rho.sum(0)

    if nderiv > 4 and dderiv > 1:
        raise NotImplementedError("Meta-GGA functional Hessians")

    if 1 < nderiv <= 4:
        rho_deriv = rho_tot[1:4, :]
    elif 4 < nderiv <= 5:
        rho_deriv = rho_tot[1:5, :]
    else:
        rho_deriv = None

    Pi_deriv = Pi[1:4, :] if nderiv_Pi > 1 else None
    xc_grid = otfnal._numint.eval_xc(otfnal.otxc, (rho_t[0, :, :],
                                                   rho_t[1, :, :]), spin=1, relativity=0, deriv=dderiv,
                                     verbose=otfnal.verbose)[:dderiv + 1]
    eot = xc_grid[0] * rho_t[:, 0, :].sum(0)
    if (weights is not None) and otfnal.verbose >= logger.DEBUG:
        nelec = rho_t[0, 0].dot(weights) + rho_t[1, 0].dot(weights)
        logger.debug(otfnal, ('MC-PDFT: Total number of electrons in (this '
                              'chunk of) the total density = %s'), nelec)
        ms = (rho_t[0, 0].dot(weights) - rho_t[1, 0].dot(weights)) / 2.0
        logger.debug(otfnal, ('MC-PDFT: Total ms = (neleca - nelecb) / 2 in '
                              '(this chunk of) the translated density = %s'), ms)
    vot = fot = None
    if dderiv > 0:
        # vrho, vsigma = xc_grid[1][:2]
        vxc = list(xc_grid[1][0].T)
        if otfnal.dens_deriv > 0:
            vxc = vxc + list(xc_grid[1][1].T)

        # vrho, vsigma, vlapl, vtau = xc_grid[1][:4]
        if otfnal.dens_deriv > 1:
            # we might get a None for one of the derivatives..
            # we get None for the laplacian derivative
            if xc_grid[1][2] is not None:
                raise NotImplementedError("laplacian translated meta-GGA functionals")

            # Here is the tau term
            vxc = vxc + list(xc_grid[1][3].T)

        vot = otfnal.jT_op(vxc, rho, Pi)
        if _unpack_vot: vot = _unpack_sigma_vector(vot,
                                                   deriv1=rho_deriv, deriv2=Pi_deriv)
    if dderiv > 1:
        # I should implement this entirely in terms of the gradient norm, since
        # that reduces the number of grid columns from 25 to 9 for t-GGA and
        # from 64 to 25 for ft-GGA (and steps around the need to "unpack"
        # fsigma and frhosigma entirely).
        fxc = _pack_fxc_ltri(xc_grid[2], otfnal.dens_deriv)
        # First pass: fxc
        fot = _jT_f_j(fxc, otfnal.jT_op, rho, Pi, rec=otfnal)
        # Second pass: translation derivatives
        fot_d_jT = otfnal.d_jT_op(vxc, rho, Pi)
        fot[:fot_d_jT.shape[0]] += fot_d_jT
    return eot, vot, fot


def unpack_vot(packed, rho, Pi):
    if rho.ndim == 2: rho = rho[:, None, :]
    if Pi.ndim == 1: Pi = Pi[None, :]
    assert (rho.shape[0] == 2)

    nderiv = rho.shape[1]
    nderiv_Pi = Pi.shape[0]

    rho_tot = rho.sum(0)
    rho_deriv = rho_tot[1:4, :] if nderiv > 1 else None
    Pi_deriv = Pi[1:4, :] if nderiv_Pi > 1 else None
    return _unpack_sigma_vector(packed, deriv1=rho_deriv, deriv2=Pi_deriv)


def _unpack_sigma_vector(packed, deriv1=None, deriv2=None):
    # For GGAs, libxc differentiates with respect to
    #   sigma[0] = nabla^2 rhoa
    #   sigma[1] = nabla rhoa . nabla rhob
    #   sigma[2] = nabla^2 rhob
    # So we have to multiply the Jacobian to obtain the requested derivatives:
    #   J[0,nabla rhoa] = 2 * nabla rhoa
    #   J[0,nabla rhob] = 0
    #   J[1,nabla rhoa] = nabla rhob
    #   J[1,nabla rhob] = nabla rhoa
    #   J[2,nabla rhoa] = 0
    #   J[2,nabla rhob] = 2 * nabla rhob
    if len(packed) > 5:
        raise RuntimeError("{} {}".format(len(packed), [p.shape for p in packed[:5]]))
    ncol1 = 1
    if deriv1 is not None and len(packed) > 2:
        ncol1 += deriv1.shape[0]
    ncol2 = 1 + 3 * int((deriv2 is not None) and len(packed) > 3)
    ngrid = packed[0].shape[-1]  # Don't assume it's an ndarray
    unp1 = np.empty((ncol1, ngrid), dtype=packed[0].dtype)
    unp2 = np.empty((ncol2, ngrid), dtype=packed[0].dtype)
    unp1[0] = packed[0]
    unp2[0] = packed[1]
    if ncol1 > 1:
        unp1[1:4] = 2 * deriv1[:3] * packed[2]
        if ncol1 > 4:
            # Deal with the tau term
            unp1[4:5] = packed[3:4]
        if ncol2 > 1:
            unp1[1:4] += deriv2 * packed[-2]
            unp2[1:4] = (2 * deriv2 * packed[-1]) + (deriv1[:3] * packed[-2])
    return unp1, unp2


def contract_vot(vot, rho, Pi):
    '''Evalute the product of unpacked vot with perturbed density, pair density, and derivatives.

        Args:
            vot : (ndarray of shape (*,ngrids), ndarray of shape (*, ngrids))
                format is ([a, ngrids], [b, ngrids]) : (vrho, vPi)
                ftGGA: a=4, b=4
                tGGA: a=4, b=1
                *tLDA: a=1, b=1
            rho : ndarray of shape (*,ngrids)
                containing density [and derivatives]
                the density contracted with vot
            Pi : ndarray with shape (*,ngrids)
                containing on-top pair density [and derivatives]
                the density contracted with vot

        Returns:
            cvot : ndarray of shape (ngrids)
                product of vot wrt (density, pair density) and their derivatives
        '''
    vrho, vPi = vot
    if rho.shape[0] == 2: rho = rho.sum(0)
    if rho.ndim == 1: rho = rho[None, :]
    if Pi.ndim == 1: Pi = Pi[None, :]

    cvot = vrho[0] * rho[0] + vPi[0] * Pi[0]
    if len(vrho) > 1:
        cvot += (vrho[1:4,:] * rho[1:4, :]).sum(0)

    if len(vPi) > 1:
        cvot += (vPi[1:4, :] * Pi[1:4, :]).sum(0)

    return cvot


def contract_fot(otfnal, fot, rho0, Pi0, rho1, Pi1, unpack=True,
                 vot_packed=None):
    r''' Evaluate the product of a packed lower-triangular matrix
    with perturbed density, pair density, and derivatives.

    Args:
        fot : ndarray of shape (*,ngrids)
            Lower-triangular matrix elements corresponding to the basis
            (rho, Pi, |drho|^2, drho'.dPi, |dPi|^2) stopping at Pi (3
            elements) for *tLDA and |drho|^2 (6 elements) for tGGA.
        rho0 : ndarray of shape (2,*,ngrids)
            containing density [and derivatives]
            the density at which fot was evaluated
        Pi0 : ndarray with shape (*,ngrids)
            containing on-top pair density [and derivatives]
            the density at which fot was evaluated
        rho1 : ndarray of shape (2,*,ngrids)
            containing density [and derivatives]
            the density contracted with fot
        Pi1 : ndarray with shape (*,ngrids)
            containing on-top pair density [and derivatives]
            the density contracted with fot

    Kwargs:
        unpack : logical
            If True, returns vot1 in unpacked shape:
                (ndarray of shape (*,ngrids),
                 ndarray of shape (*,ngrids))
            corresponding to (density, pair density) and their
            derivatives. This requires vot_packed for *tGGA functionals
            Otherwise, returns vot1 in packed shape:
                (rho, Pi, |rho'|^2, rho'.Pi', |Pi'|^2)
            stopping at Pi (3 elements) for *tLDA and |rho'|^2 (6
            elements) for tGGA.
        vot_packed : ndarray of shape (*,ngrids)
            Vector elements corresponding to the basis
            (rho, Pi, |drho|^2, drho'.dPi, |dPi|^2) stopping at Pi (2
            elements) for *tLDA and |drho|^2 (3 elements) for tGGA.
            Required if unpack == True for *tGGA functionals
            (because vot_|drho|^2 contributes to fot_rho',rho', etc.)

    Returns:
        vot1 : (ndarray of shape (*,ngrids),
                ndarray of shape (*,ngrids))
            product of fot wrt (density, pair density)
            and their derivatives
    '''
    if rho0.shape[0] == 2: rho0 = rho0.sum(0)  # Never has exactly 1 derivative
    if rho0.ndim == 1: rho0 = rho0[None, :]
    if Pi0.ndim == 1: Pi0 = Pi0[None, :]
    if rho1.shape[0] == 2: rho1 = rho1.sum(0)  # Never has exactly 1 derivative
    if rho1.ndim == 1: rho1 = rho1[None, :]
    if Pi1.ndim == 1: Pi1 = Pi1[None, :]

    ngrids = fot[0].shape[-1]
    vrho1 = np.zeros(ngrids, dtype=fot[0].dtype)
    vPi1 = np.zeros(ngrids, dtype=fot[2].dtype)
    vrho1, vPi1 = np.zeros_like(rho1), np.zeros_like(Pi1)

    # TODO: dspmv implementation
    vrho1 = fot[0] * rho1[0] + fot[1] * Pi1[0]
    vPi1 = fot[2] * Pi1[0] + fot[1] * rho1[0]

    rho0p = Pi0p = rho1p = Pi1p = None
    v1 = [vrho1, vPi1]
    if len(fot) > 3:
        srr = 2 * (rho0[1:4, :] * rho1[1:4, :]).sum(0)
        vrho1 += fot[3] * srr
        vPi1 += fot[4] * srr
        vrr = fot[3] * rho1[0] + fot[4] * Pi1[0] + fot[5] * srr
        rho0p = rho0[1:4]
        rho1p = rho1[1:4]
        v1 = [vrho1, vPi1, vrr]
    if len(fot) > 6:
        srP = ((rho0[1:4, :] * Pi1[1:4, :]).sum(0)
               + (rho1[1:4, :] * Pi0[1:4, :]).sum(0))
        sPP = 2 * (Pi0[1:4, :] * Pi1[1:4, :]).sum(0)
        vrho1 += fot[6] * srP + fot[10] * sPP
        vPi1 += fot[7] * srP + fot[11] * sPP
        vrr += fot[8] * srP + fot[12] * sPP
        vrP = (fot[6] * rho1[0] + fot[7] * Pi1[0]
               + fot[8] * srr + fot[9] * srP + fot[13] * sPP)
        vPP = (fot[10] * rho1[0] + fot[11] * Pi1[0]
               + fot[12] * srr + fot[13] * srP + fot[14] * sPP)
        Pi0p = Pi0[1:4]
        Pi1p = Pi1[1:4]
        v1 = [vrho1, vPi1, vrr, vrP, vPP]

    ggrad = (rho0p is not None) or (Pi0p is not None)
    if unpack:
        vrho1, vPi1 = _unpack_sigma_vector(v1, rho0p, Pi0p)
        if ggrad:
            if vot_packed is None:
                raise RuntimeError("Cannot evaluate fot.x in terms of "
                                    "unpacked density gradients without vot_packed")
            vrho2, vPi2 = _unpack_sigma_vector(vot_packed, rho1p, Pi1p)
            if vrho1.shape[0] > 1: vrho1[1:4, :] += vrho2[1:4, :]
            if vPi1.shape[0] > 1:  vPi1[1:4, :] += vPi2[1:4, :]
        v1 = (vrho1, vPi1)

    return v1


def _jT_f_j(frr, jT_op, *args, **kwargs):
    r''' Apply a jacobian function taking *args to the lower-triangular
    second-derivative array frr'''
    nel = len(frr)
    nr = int(round(np.sqrt(1 + 8 * nel) - 1)) // 2
    rec = kwargs.get('rec', None)
    ngrids = frr[0].shape[-1]

    # build square-matrix index array to address packed matrix frr w/o copying
    ltri_ix = np.tril_indices(nr)
    idx_arr = np.zeros((nr, nr), dtype=np.int32)
    idx_arr[ltri_ix] = range(nel)
    idx_arr += idx_arr.T
    diag_ix = np.diag_indices(nr)
    idx_arr[diag_ix] = idx_arr[diag_ix] // 2

    # first pass: jT . frr -> fcr
    fcr = np.stack([jT_op([frr[i] for i in ix_row], *args)
                    for ix_row in idx_arr], axis=1)

    # second pass. fcr is a rectangular matrix (unavoidably)
    nc = fcr.shape[0]
    if getattr(rec, 'verbose', 0) < logger.DEBUG:
        fcc = np.empty((nc * (nc + 1) // 2, ngrids), dtype=fcr.dtype)
        i = 0
        for ix_row, fc_row in enumerate(fcr):
            di = ix_row + 1
            j = i + di
            fcc[i:j] = jT_op(fc_row, *args)[:di]
            i = j
    else:
        fcc = np.empty((nc, nc, ngrids), dtype=fcr.dtype)
        for fcc_row, fcr_row in zip(fcc, fcr):
            fcc_row[:] = jT_op(fcr_row, *args)
        for i in range(1, nc):
            for j in range(i):
                scale = (fcc[i, j] + fcc[j, i]) / 2
                scale[scale == 0] = 1
                logger.debug(rec, 'MC-PDFT jT_f_j symmetry check %d,%d: %e',
                             i, j, linalg.norm((fcc[i, j] - fcc[j, i]) / scale))
        ltri_ix = np.tril_indices(nc)
        fcc = fcc[ltri_ix]

    return fcc


def _gentLDA_jT_op(x, rho, Pi, R, zeta):
    # On a grid, multiply the transpose of the Jacobian
    #     d(trhoa,trhob) [fictitous densities]
    # J = ______________
    #     d(rho,Pi)      [real densities]
    # by a vector x_(trhoa,trhob)
    ngrid = rho.shape[-1]
    if R.ndim > 1: R = R[0]

    # ab -> cs coordinate transformation
    xc = (x[0] + x[1]) / 2.0
    xm = (x[0] - x[1]) / 2.0

    # Charge sector has no explicit rho denominator
    # and so does not require indexing to avoid
    # division by zero
    jTx = np.zeros((2, ngrid), dtype=x[0].dtype)
    jTx[0] = xc + xm * (zeta[0] - (2 * R * zeta[1]))

    # Spin sector has a rho denominator
    idx = (rho[0] > 1e-15)
    zeta = zeta[1, idx]
    rho = rho[0, idx]
    xm = xm[idx]
    jTx[1, idx] = 4 * xm * zeta / rho

    return jTx


def _tGGA_jT_op(x, rho, Pi, R, zeta):
    # On a grid, multiply the transpose of the Jacobian
    #     d(trho?'.trho?') [fictitous density gradients]
    # J = ______________
    #     d(rho,Pi,|rho'|) [real densities and gradients]
    # by a vector x_(|trho*'|^2) in the context of tGGAs
    ngrid = rho.shape[-1]
    jTx = np.zeros((3, ngrid), dtype=x[0].dtype)
    if R.ndim > 1: R = R[0]

    # ab -> cs coordinate transformation
    xcc = (x[2] + x[4] + x[3]) / 4.0
    xcm = (x[2] - x[4]) / 2.0
    xmm = (x[2] + x[4] - x[3]) / 4.0

    # Gradient-gradient sector
    idx = (zeta[0]!=1)
    jTx[2] = x[2]
    jTx[2,idx] = (xcc + xcm * zeta[0] + xmm * zeta[0] * zeta[0])[idx]

    # Finite-precision safety

    # Density-gradient sector
    idx = (rho[0] > 1e-15)
    sigma_fac = ((rho[1:4].conj() * rho[1:4]).sum(0) * zeta[1])
    sigma_fac = ((xcm + 2 * zeta[0] * xmm) * sigma_fac)[idx]
    rho = rho[0, idx]
    R = R[idx]
    sigma_fac = -2 * sigma_fac / rho
    jTx[0, idx] = R * sigma_fac
    jTx[1, idx] = -2 * sigma_fac / rho

    return jTx


def _tmetaGGA_jT_op(x, rho, Pi, R, zeta):
    # output ordering is
    # ordering: rho, Pi, |rho'|^2, tau
    ngrid = rho.shape[-1]
    jTx = np.zeros((4, ngrid), dtype=x[0].dtype)
    if R.ndim > 1:
        R = R[0]

    # ab -> cs coordinate transformation
    xc = (x[5] + x[6]) / 2.0
    xm = (x[5] - x[6]) / 2.0

    # easy part
    jTx[3] = xc + zeta[0] * xm

    tau_lapl_factor = zeta[1] * rho[4] * xm
    idx = rho[0] > 1e-15
    rho = rho[0, idx]
    R = R[idx]

    tau_lapl_factor = 2 * tau_lapl_factor[idx] / rho

    jTx[0, idx] = -R * tau_lapl_factor
    jTx[1, idx] = 2 * tau_lapl_factor / rho

    return jTx


def _tGGA_jT_op_m2z(x, rho, zeta, srr):
    # cs -> rho,zeta step of _tGGA_jT_op above
    # unused; for contemplative purposes only
    jTx = np.empty_like(x)
    jTx[0] = x[0] + zeta[0] * x[1]
    jTx[1] = x[1] * rho + (x[3] + 2 * zeta[0] * x[4]) * srr
    jTx[2] = x[2] + x[3] * zeta[0] + x[4] * zeta[0] * zeta[0]
    jTx[3] = 0
    jTx[4] = 0
    return jTx


def _ftGGA_jT_op_m2z(x, rho, zeta, srz, szz):
    # cs -> rho,zeta step of _ftGGA_jT_op below
    jTx = np.empty_like(x)
    jTx[0] = 2 * x[4] * (zeta[0] * srz + rho * szz) + x[3] * srz
    jTx[1] = 2 * x[4] * rho * srz
    jTx[2] = 0
    jTx[3] = (x[3] + 2 * x[4] * zeta[0]) * rho
    jTx[4] = x[4] * rho * rho
    return jTx


def _ftGGA_jT_op_z2R(x, zeta, srP, sPP):
    # rho,zeta -> rho,R step of _ftGGA_jT_op below
    jTx = np.empty_like(x)
    jTx[0] = x[0]
    jTx[1] = (x[1] * zeta[1] + x[3] * srP * zeta[2] +
              2 * x[4] * sPP * zeta[1] * zeta[2])
    jTx[2] = x[2]
    jTx[3] = x[3] * zeta[1]
    jTx[4] = x[4] * zeta[1] * zeta[1]
    return jTx


def _ftGGA_jT_op_R2Pi(x, rho, R, srr, srP, sPP):
    # rho,R -> rho,Pi step of _ftGGA_jT_op below
    if rho.ndim > 1: rho = rho[0]
    if R.ndim > 1: R = R[0]
    jTx = np.empty_like(x)
    ri = np.empty_like(x)
    ri[0, :] = 0.0
    idx = rho > 1e-15
    ri[0, idx] = 1.0 / rho[idx]
    for i in range(4):
        ri[i + 1] = ri[i] * ri[0]

    jTx[0] = (x[0] - 2 * R * x[1] * ri[0]
              + x[3] * (6 * R * ri[1] * srr - 8 * srP * ri[2])
              + x[4] * (-24 * R * R * ri[2] * srr + 80 * R * ri[3] * srP
                        - 64 * ri[4] * sPP))
    jTx[1] = (4 * x[1] * ri[1] - 8 * x[3] * ri[2] * srr
              + x[4] * (32 * R * ri[3] * srr - 64 * ri[4] * srP))
    jTx[2] = x[2] - 2 * R * x[3] * ri[0] + 4 * x[4] * R * R * ri[1]
    jTx[3] = 4 * x[3] * ri[1] - 16 * x[4] * R * ri[2]
    jTx[4] = 16 * x[4] * ri[3]
    return jTx


def _ftGGA_jT_op(x, rho, Pi, R, zeta):
    # On a grid, evaluate the contribution to the matrix-vector product
    # of the transpose of the Jacobian
    #     d(trho?'.trho?') [fictitous density gradients]
    # J = ______________
    #     d(rho,Pi,|rho'|) [real densities and gradients]
    # with a vector x_(|trho*'|^2), which is present in ftGGAs and
    # missing in tGGAs
    ngrid = rho.shape[-1]
    jTx = np.zeros((5, ngrid), dtype=x[0].dtype)

    # ab -> cs step
    jTx[2] = (x[2] + x[4] + x[3]) / 4.0
    jTx[3] = (x[2] - x[4]) / 2.0
    jTx[4] = (x[2] + x[4] - x[3]) / 4.0
    x = jTx

    # Intermediates
    srr = (rho[1:4, :] * rho[1:4, :]).sum(0)
    srP = (rho[1:4, :] * R[1:4, :]).sum(0)
    sPP = (R[1:4, :] * R[1:4, :]).sum(0)
    srz = srP * zeta[1]
    szz = sPP * zeta[1] * zeta[1]

    # cs -> rho,zeta step
    x = _ftGGA_jT_op_m2z(x, rho[0], zeta, srz, szz)

    # rho,zeta -> rho,R step
    x = _ftGGA_jT_op_z2R(x, zeta, srP, sPP)

    # rho,R -> rho,Pi step
    srP = (rho[1:4, :] * Pi[1:4, :]).sum(0)
    sPP = (Pi[1:4, :] * Pi[1:4, :]).sum(0)
    jTx = _ftGGA_jT_op_R2Pi(x, rho, R, srr, srP, sPP)

    return jTx


def _gentLDA_d_jT_op(x, rho, Pi, R, zeta):
    # On a grid, differentiate once the Jacobian
    #     d(trhoa,trhob) [fictitous densities]
    # J = ______________
    #     d(rho,Pi)      [real densities]
    # and multiply by x_(trhoa,trhob) so as to compute the nonlinear-
    # translation contribution to the second functional derivatives of
    # the on-top energy in tLDA and ftLDA.
    rho = rho[0]
    Pi = Pi[0]
    R = R[0]
    ngrid = rho.shape[-1]
    f = np.zeros((3, ngrid), dtype=x[0].dtype)

    # ab -> cs
    xm = (x[0] - x[1]) / 2.0

    # Indexing
    idx = rho > 1e-15
    rho = rho[idx]
    Pi = Pi[idx]
    xm = xm[idx]
    R = R[idx]
    zeta = zeta[:, idx]

    # Intermediates
    # R = otfnal.get_ratio (Pi, rho/2)
    # zeta = otfnal.get_zeta (R, fn_deriv=2)[1:]
    xmw = 2 * xm / rho
    z1 = xmw * (zeta[1] + 2 * R * zeta[2])

    # without further ceremony
    f[0, idx] = R * z1
    f[1, idx] = -2 * z1 / rho
    f[2, idx] = xmw * 8 * zeta[2] / rho / rho

    return f


def _tGGA_d_jT_op(x, rho, Pi, R, zeta):
    # On a grid, differentiate once the Jacobian
    #     d(trho?'.trho?') [fictitous density gradients]
    # J = ______________
    #     d(rho,Pi,|rho'|) [real densities and gradients]
    # and multiply by x_(trho?'.trho?') so as to compute the nonlinear-
    # translation contribution to the second functional derivatives of
    # the on-top energy in tGGAs.

    # Generates contributions to the first five elements
    # of the lower-triangular packed Hessian
    ngrid = rho.shape[-1]
    f = np.zeros((5, ngrid), dtype=x[0].dtype)

    # Indexing
    idx = rho[0] > 1e-15
    rho = rho[0:4, idx]
    Pi = Pi[0:1, idx]
    x = [xi[idx] for xi in x]
    R = R[0, idx]
    zeta = zeta[:, idx]

    # ab -> cs
    xcm = (x[2] - x[4]) / 2.0
    xmm = (x[2] + x[4] - x[3]) / 4.0

    # Intermediates
    sigma = (rho[1:4] * rho[1:4]).sum(0)
    rho = rho[0]
    rho2 = rho * rho
    rho3 = rho2 * rho
    rho4 = rho3 * rho

    # coefficient of dsigma dz
    xcm += 2 * zeta[0] * xmm
    f[3, idx] = -2 * xcm * R * zeta[1] / rho
    f[4, idx] = 4 * xcm * zeta[1] / rho2

    # coefficient of d^2 z
    xcm *= sigma
    f[0, idx] = 2 * xcm * R * (3 * zeta[1] + 2 * R * zeta[2]) / rho2
    f[1, idx] = -8 * xcm * (zeta[1] + R * zeta[2]) / rho3
    f[2, idx] = 16 * xcm * zeta[2] / rho4

    # coefficient of dz dz
    xmm *= 8 * sigma * zeta[1] * zeta[1] / rho2
    f[0, idx] += xmm * R * R
    f[1, idx] -= 2 * xmm * R / rho
    f[2, idx] += 4 * xmm / rho2

    return f


#   r,r
#   1,r   1,1
#   srr,r srr,1 srr,srr
#   sr1,r sr1,1 sr1,srr sr1,sr1
#   s11,r s11,1 s11,srr s11,sr1 s11,s11

def _ftGGA_d_jT_op_m2z(v, rho, zeta, srz, szz):
    # srm += srz*r
    # smm += 2srz*r*z + szz*r*r
    # 0  : r, r
    # 1  : r, z
    # 6  : srz, r
    # 7  : srz, z
    # 10 : szz, r
    ngrids = v.shape[1]
    f = np.zeros((15, ngrids), dtype=v.dtype)
    f[0] = 2 * v[4] * szz
    f[1] = 2 * v[4] * srz
    f[6] = v[3] + 2 * v[4] * zeta[0]
    f[7] = 2 * v[4] * rho
    f[10] = 2 * v[4] * rho
    assert (tuple(f[0].shape) == tuple(f[1].shape))
    assert (tuple(f[0].shape) == tuple(f[6].shape))
    assert (tuple(f[0].shape) == tuple(f[10].shape))
    return f


def _ftGGA_d_jT_op_z2R(v, zeta, srP, sPP):
    # z = z[0]
    # srz = srP*z[1]
    # szz = sPP*z[1]**2
    # 2  : P, P
    # 7  : srP, P
    # 11 : sPP, P
    ngrids = v.shape[1]
    f = np.zeros((15, ngrids), dtype=v.dtype)
    f[2] = 2 * v[4] * sPP * (zeta[3] * zeta[1] + zeta[2] * zeta[2])
    f[2] += v[1] * zeta[2] + v[3] * srP * zeta[3]
    f[7] = v[3] * zeta[2]
    f[11] = 2 * v[4] * zeta[1] * zeta[2]
    assert (tuple(f[2].shape) == tuple(f[7].shape))
    assert (tuple(f[2].shape) == tuple(f[11].shape))
    return f


def _ftGGA_d_jT_op_R2Pi(v, rho, Pi, srr, srP, sPP):
    # R = 4Pi/(r**2) = Pi*d[0]
    # srR = srP*d[0] + srr*Pi*d[1]
    # sRR = sPP*d[0]**2 + 2*Pi*d[1]*srP*d[0] + srr*(Pi*d[1])**2
    #        d^n(d[0])
    # d[n] = ---------
    #          dr^n
    ngrids = v.shape[-1]
    f = np.zeros((15, ngrids), dtype=v.dtype)
    d = np.zeros((4, ngrids), dtype=v.dtype)
    idx = np.abs(rho) > 1e-15
    d[0, idx] = 4 / rho[idx] / rho[idx]
    d[1, idx] = -2 * d[0, idx] / rho[idx]
    d[2, idx] = -3 * d[1, idx] / rho[idx]
    d[3, idx] = -4 * d[2, idx] / rho[idx]
    # rho, rho
    f[0] = v[1] * Pi * d[2]
    f[0] += v[3] * (srP * d[2] + srr * Pi * d[3])
    f[0] += 2 * v[4] * sPP * (d[2] * d[0] + d[1] * d[1])
    f[0] += 2 * v[4] * srP * Pi * (3 * d[2] * d[1] + d[3] * d[0])
    f[0] += 2 * v[4] * srr * Pi * Pi * (d[3] * d[1] + d[2] * d[2])
    # rho, Pi
    f[1] = v[1] * d[1] + v[3] * srr * d[2]
    f[1] += 2 * v[4] * srP * (d[2] * d[0] + d[1] * d[1])
    f[1] += 4 * v[4] * srr * Pi * d[2] * d[1]
    # Pi, Pi
    f[2] = 2 * v[4] * srr * d[1] * d[1]
    # rho, rr
    f[3] = v[3] * Pi * d[2] + 2 * v[4] * Pi * Pi * d[2] * d[1]
    # Pi, rr
    f[4] = v[3] * d[1] + 2 * v[4] * Pi * d[1] * d[1]
    # rho, rP
    f[6] = v[3] * d[1] + 2 * v[4] * Pi * (d[2] * d[0] + d[1] * d[1])
    # Pi, rP
    f[7] = 2 * v[4] * d[0] * d[1]
    # rho, PP
    f[10] = 2 * v[4] * d[0] * d[1]
    for row in f[1:]:
        if hasattr(row, 'shape'):
            assert (tuple(row.shape) == tuple(f[0].shape))
    return f


def _ftGGA_d_jT_op(v, rho, Pi, R, zeta):
    # raise NotImplementedError ("Second density derivatives for fully-"
    #    "translated GGA functionals")
    # Generates contributions to the first five elements,
    # then 6,7, then 10,11
    # of the lower-triangular packed Hessian
    # (I.E., no double gradient derivatives)
    # for the terms added in the fully-translated extension of tGGA

    # ab -> cs
    vcm = (v[2] - v[4]) / 2.0
    vmm = (v[2] + v[4] - v[3]) / 4.0
    v[3] = vcm
    v[4] = vmm
    v = np.asarray(v)

    # Intermediates
    srr = (rho[1:4, :] * rho[1:4, :]).sum(0)
    srP = (rho[1:4, :] * R[1:4, :]).sum(0)
    sPP = (R[1:4, :] * R[1:4, :]).sum(0)
    srz = srP * zeta[1]
    szz = sPP * zeta[1] * zeta[1]

    # cs -> rho, zeta
    f = _ftGGA_d_jT_op_m2z(v, rho[0], zeta, srz, szz)
    v = _ftGGA_jT_op_m2z(v, rho[0], zeta, srz, szz)

    # rho, zeta -> rho, R
    # The for loops here are because I'm guessing that repeated
    # initialization of large arrays to zero is slower than a short,
    # shallow Python loop
    f = _jT_f_j(f, _ftGGA_jT_op_z2R, zeta, srP, sPP)
    f += _ftGGA_d_jT_op_z2R(v, zeta, srP, sPP)
    v = _ftGGA_jT_op_z2R(v, zeta, srP, sPP)

    # rho, R -> rho, Pi
    srP = (rho[1:4, :] * Pi[1:4, :]).sum(0)
    sPP = (Pi[1:4, :] * Pi[1:4, :]).sum(0)
    f = _jT_f_j(f, _ftGGA_jT_op_R2Pi, rho, R, srr, srP, sPP)
    f += _ftGGA_d_jT_op_R2Pi(v, rho[0], Pi[0], srr, srP, sPP)

    return f



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


'''
RMP2
'''


import numpy
from pyscf import gto
from pyscf import lib
from pyscf.lib import logger
from pyscf import ao2mo
from pyscf.ao2mo import _ao2mo
from pyscf import __config__

WITH_T2 = getattr(__config__, 'mp_mp2_with_t2', True)


def kernel(mp, mo_energy=None, mo_coeff=None, eris=None, with_t2=WITH_T2, verbose=None):
    if mo_energy is not None or mo_coeff is not None:
        # For backward compatibility.  In pyscf-1.4 or earlier, mp.frozen is
        # not supported when mo_energy or mo_coeff is given.
        assert (mp.frozen == 0 or mp.frozen is None)

    if eris is None:
        eris = mp.ao2mo(mo_coeff)

    if mo_energy is None:
        mo_energy = eris.mo_energy

    nocc = mp.nocc
    nvir = mp.nmo - nocc
    eia = mo_energy[:nocc,None] - mo_energy[None,nocc:]

    if with_t2:
        t2 = numpy.empty((nocc,nocc,nvir,nvir), dtype=eris.ovov.dtype)
    else:
        t2 = None

    emp2_ss = emp2_os = 0
    for i in range(nocc):
        if isinstance(eris.ovov, numpy.ndarray) and eris.ovov.ndim == 4:
            # When mf._eri is a custom integrals with the shape (n,n,n,n), the
            # ovov integrals might be in a 4-index tensor.
            gi = eris.ovov[i]
        else:
            gi = numpy.asarray(eris.ovov[i*nvir:(i+1)*nvir])

        gi = gi.reshape(nvir,nocc,nvir).transpose(1,0,2)
        t2i = gi.conj()/lib.direct_sum('jb+a->jba', eia, eia[i])
        edi = numpy.einsum('jab,jab', t2i, gi) * 2
        exi = -numpy.einsum('jab,jba', t2i, gi)
        emp2_ss += edi*0.5 + exi
        emp2_os += edi*0.5
        if with_t2:
            t2[i] = t2i

    emp2_ss = emp2_ss.real
    emp2_os = emp2_os.real
    emp2 = lib.tag_array(emp2_ss+emp2_os, e_corr_ss=emp2_ss, e_corr_os=emp2_os)

    return emp2.real, t2


# Iteratively solve MP2 if non-canonical HF is provided
def _iterative_kernel(mp, eris, verbose=None):
    cput1 = cput0 = (logger.process_clock(), logger.perf_counter())
    log = logger.new_logger(mp, verbose)

    emp2, t2 = mp.init_amps(eris=eris)
    log.info('Init E(MP2) = %.15g', emp2)

    adiis = lib.diis.DIIS(mp)

    conv = False
    for istep in range(mp.max_cycle):
        t2new = mp.update_amps(t2, eris)

        if isinstance(t2new, numpy.ndarray):
            normt = numpy.linalg.norm(t2new - t2)
            t2 = None
            t2new = adiis.update(t2new)
        else: # UMP2
            normt = numpy.linalg.norm([numpy.linalg.norm(t2new[i] - t2[i])
                                       for i in range(3)])
            t2 = None
            t2shape = [x.shape for x in t2new]
            t2new = numpy.hstack([x.ravel() for x in t2new])
            t2new = adiis.update(t2new)
            t2new = lib.split_reshape(t2new, t2shape)

        t2, t2new = t2new, None
        emp2, e_last = mp.energy(t2, eris), emp2
        log.info('cycle = %d  E_corr(MP2) = %.15g  dE = %.9g  norm(t2) = %.6g',
                 istep+1, emp2, emp2 - e_last, normt)
        cput1 = log.timer('MP2 iter', *cput1)
        if abs(emp2-e_last) < mp.conv_tol and normt < mp.conv_tol_normt:
            conv = True
            break
    log.timer('MP2', *cput0)
    return conv, emp2, t2

def energy(mp, t2, eris):
    '''MP2 energy'''
    nocc, nvir = t2.shape[1:3]
    eris_ovov = numpy.asarray(eris.ovov).reshape(nocc,nvir,nocc,nvir)
    ed = numpy.einsum('ijab,iajb', t2, eris_ovov) * 2
    ex = -numpy.einsum('ijab,ibja', t2, eris_ovov)
    emp2_ss = (ed*0.5 + ex).real
    emp2_os = ed.real*0.5
    emp2 = lib.tag_array(emp2_ss+emp2_os, e_corr_ss=emp2_ss, e_corr_os=emp2_os)
    return emp2

def update_amps(mp, t2, eris):
    '''Update non-canonical MP2 amplitudes'''
    #assert (isinstance(eris, _ChemistsERIs))
    nocc, nvir = t2.shape[1:3]
    fock = eris.fock
    mo_e_o = eris.mo_energy[:nocc]
    mo_e_v = eris.mo_energy[nocc:] + mp.level_shift

    foo = fock[:nocc,:nocc] - numpy.diag(mo_e_o)
    fvv = fock[nocc:,nocc:] - numpy.diag(mo_e_v)
    t2new  = lib.einsum('ijac,bc->ijab', t2, fvv)
    t2new -= lib.einsum('ki,kjab->ijab', foo, t2)
    t2new = t2new + t2new.transpose(1,0,3,2)

    eris_ovov = numpy.asarray(eris.ovov).reshape(nocc,nvir,nocc,nvir)
    t2new += eris_ovov.conj().transpose(0,2,1,3)
    eris_ovov = None

    eia = mo_e_o[:,None] - mo_e_v
    t2new /= lib.direct_sum('ia,jb->ijab', eia, eia)
    return t2new


def make_rdm1(mp, t2=None, eris=None, ao_repr=False, with_frozen=True):
    r'''Spin-traced one-particle density matrix.
    The occupied-virtual orbital response is not included.

    dm1[p,q] = <q_alpha^\dagger p_alpha> + <q_beta^\dagger p_beta>

    The convention of 1-pdm is based on McWeeney's book, Eq (5.4.20).
    The contraction between 1-particle Hamiltonian and rdm1 is
    E = einsum('pq,qp', h1, rdm1)

    Kwargs:
        ao_repr : boolean
            Whether to transform 1-particle density matrix to AO
            representation.
    '''
    from pyscf.cc import ccsd_rdm
    doo, dvv = _gamma1_intermediates(mp, t2, eris)
    nocc = doo.shape[0]
    nvir = dvv.shape[0]
    dov = numpy.zeros((nocc,nvir), dtype=doo.dtype)
    dvo = dov.T
    return ccsd_rdm._make_rdm1(mp, (doo, dov, dvo, dvv), with_frozen=with_frozen,
                               ao_repr=ao_repr)

def _gamma1_intermediates(mp, t2=None, eris=None):
    if t2 is None: t2 = mp.t2
    nmo = mp.nmo
    nocc = mp.nocc
    nvir = nmo - nocc
    if t2 is None:
        if eris is None:
            eris = mp.ao2mo()
        mo_energy = eris.mo_energy
        eia = mo_energy[:nocc,None] - mo_energy[None,nocc:]
        dtype = eris.ovov.dtype
    else:
        dtype = t2.dtype

    dm1occ = numpy.zeros((nocc,nocc), dtype=dtype)
    dm1vir = numpy.zeros((nvir,nvir), dtype=dtype)
    for i in range(nocc):
        if t2 is None:
            gi = numpy.asarray(eris.ovov[i*nvir:(i+1)*nvir])
            gi = gi.reshape(nvir,nocc,nvir).transpose(1,0,2)
            t2i = gi.conj()/lib.direct_sum('jb+a->jba', eia, eia[i])
        else:
            t2i = t2[i]
        l2i = t2i.conj()
        dm1vir += lib.einsum('jca,jcb->ba', l2i, t2i) * 2 \
                - lib.einsum('jca,jbc->ba', l2i, t2i)
        dm1occ += lib.einsum('iab,jab->ij', l2i, t2i) * 2 \
                - lib.einsum('iab,jba->ij', l2i, t2i)
    return -dm1occ, dm1vir


def _mo_splitter(mp):
    maskact = mp.get_frozen_mask()
    maskocc = mp.mo_occ>1e-6
    masks = [
        maskocc  & ~maskact,    # frz occ
        maskocc  &  maskact,    # act occ
        ~maskocc &  maskact,    # act vir
        ~maskocc & ~maskact,    # frz vir
    ]
    return masks

def make_fno(mp, thresh=1e-6, pct_occ=None, nvir_act=None, t2=None):
    r'''
    Frozen natural orbitals

    Attributes:
        thresh : float
            Threshold on NO occupation numbers. Default is 1e-6 (very conservative).
        pct_occ : float
            Percentage of total occupation number. Default is None. If present, overrides `thresh`.
        nvir_act : int
            Number of virtual NOs to keep. Default is None. If present, overrides `thresh` and `pct_occ`.

    Returns:
        frozen : list or ndarray
            List of orbitals to freeze
        no_coeff : ndarray
            Semicanonical NO coefficients in the AO basis
    '''
    mf = mp._scf
    dm = mp.make_rdm1(t2=t2, with_frozen=False)

    nmo = mp.nmo
    nocc = mp.nocc
    nvir = nmo - nocc
    n,v = numpy.linalg.eigh(dm[nocc:,nocc:])
    idx = numpy.argsort(n)[::-1]
    n,v = n[idx], v[:,idx]

    if nvir_act is None:
        if pct_occ is None:
            nvir_keep = numpy.count_nonzero(n>thresh)
        else:
            cumsum = numpy.cumsum(n/numpy.sum(n))
            logger.debug(mp, 'Sum(pct_occ): %s', cumsum)
            nvir_keep = numpy.count_nonzero(
                [c <= pct_occ or numpy.isclose(c, pct_occ) for c in cumsum])
    else:
        nvir_keep = min(nvir, nvir_act)

    masks = _mo_splitter(mp)
    moeoccfrz0, moeocc, moevir, moevirfrz0 = [mf.mo_energy[m] for m in masks]
    orboccfrz0, orbocc, orbvir, orbvirfrz0 = [mf.mo_coeff[:,m] for m in masks]

    fvv = numpy.diag(moevir)
    fvv_no = numpy.dot(v.T, numpy.dot(fvv, v))
    _, v_canon = numpy.linalg.eigh(fvv_no[:nvir_keep,:nvir_keep])

    orbviract = numpy.dot(orbvir, numpy.dot(v[:,:nvir_keep], v_canon))
    orbvirfrz = numpy.dot(orbvir, v[:,nvir_keep:])
    no_comp = (orboccfrz0, orbocc, orbviract, orbvirfrz, orbvirfrz0)
    no_coeff = numpy.hstack(no_comp)
    nocc_loc = numpy.cumsum([0]+[x.shape[1] for x in no_comp]).astype(int)
    no_frozen = numpy.hstack((numpy.arange(nocc_loc[0], nocc_loc[1]),
                              numpy.arange(nocc_loc[3], nocc_loc[5]))).astype(int)

    return no_frozen, no_coeff


def make_rdm2(mp, t2=None, eris=None, ao_repr=False):
    r'''
    Spin-traced two-particle density matrix in MO basis

    dm2[p,q,r,s] = \sum_{sigma,tau} <p_sigma^\dagger r_tau^\dagger s_tau q_sigma>

    Note the contraction between ERIs (in Chemist's notation) and rdm2 is
    E = einsum('pqrs,pqrs', eri, rdm2)
    '''
    if t2 is None: t2 = mp.t2
    nmo = nmo0 = mp.nmo
    nocc = nocc0 = mp.nocc
    nvir = nmo - nocc
    if t2 is None:
        if eris is None:
            eris = mp.ao2mo()
        mo_energy = eris.mo_energy
        eia = mo_energy[:nocc,None] - mo_energy[None,nocc:]

    if mp.frozen is not None:
        nmo0 = mp.mo_occ.size
        nocc0 = numpy.count_nonzero(mp.mo_occ > 0)
        moidx = get_frozen_mask(mp)
        oidx = numpy.where(moidx & (mp.mo_occ > 0))[0]
        vidx = numpy.where(moidx & (mp.mo_occ ==0))[0]
    else:
        moidx = oidx = vidx = None

    dm1 = make_rdm1(mp, t2, eris)
    dm1[numpy.diag_indices(nocc0)] -= 2

    dm2 = numpy.zeros((nmo0,nmo0,nmo0,nmo0), dtype=dm1.dtype) # Chemist notation
    #dm2[:nocc,nocc:,:nocc,nocc:] = t2.transpose(0,3,1,2)*2 - t2.transpose(0,2,1,3)
    #dm2[nocc:,:nocc,nocc:,:nocc] = t2.transpose(3,0,2,1)*2 - t2.transpose(2,0,3,1)
    for i in range(nocc):
        if t2 is None:
            gi = numpy.asarray(eris.ovov[i*nvir:(i+1)*nvir])
            gi = gi.reshape(nvir,nocc,nvir).transpose(1,0,2)
            t2i = gi.conj()/lib.direct_sum('jb+a->jba', eia, eia[i])
        else:
            t2i = t2[i]
        # dm2 was computed as dm2[p,q,r,s] = < p^\dagger r^\dagger s q > in the
        # above. Transposing it so that it be contracted with ERIs (in Chemist's
        # notation):
        #   E = einsum('pqrs,pqrs', eri, rdm2)
        dovov = t2i.transpose(1,0,2)*2 - t2i.transpose(2,0,1)
        dovov *= 2
        if moidx is None:
            dm2[i,nocc:,:nocc,nocc:] = dovov
            dm2[nocc:,i,nocc:,:nocc] = dovov.conj().transpose(0,2,1)
        else:
            dm2[oidx[i],vidx[:,None,None],oidx[:,None],vidx] = dovov
            dm2[vidx[:,None,None],oidx[i],vidx[:,None],oidx] = dovov.conj().transpose(0,2,1)

    # Be careful with convention of dm1 and dm2
    #   dm1[q,p] = <p^\dagger q>
    #   dm2[p,q,r,s] = < p^\dagger r^\dagger s q >
    #   E = einsum('pq,qp', h1, dm1) + .5 * einsum('pqrs,pqrs', eri, dm2)
    # When adding dm1 contribution, dm1 subscripts need to be flipped
    for i in range(nocc0):
        dm2[i,i,:,:] += dm1.T * 2
        dm2[:,:,i,i] += dm1.T * 2
        dm2[:,i,i,:] -= dm1.T
        dm2[i,:,:,i] -= dm1

    for i in range(nocc0):
        for j in range(nocc0):
            dm2[i,i,j,j] += 4
            dm2[i,j,j,i] -= 2

    if ao_repr:
        from pyscf.cc import ccsd_rdm
        dm2 = ccsd_rdm._rdm2_mo2ao(dm2, mp.mo_coeff)
    return dm2


def get_nocc(mp):
    if mp._nocc is not None:
        return mp._nocc
    elif mp.frozen is None:
        nocc = numpy.count_nonzero(mp.mo_occ > 0)
        assert (nocc > 0)
        return nocc
    elif isinstance(mp.frozen, (int, numpy.integer)):
        nocc = numpy.count_nonzero(mp.mo_occ > 0) - mp.frozen
        assert (nocc > 0)
        return nocc
    elif hasattr(mp.frozen, '__len__'):
        occ_idx = mp.mo_occ > 0
        occ_idx[list(mp.frozen)] = False
        nocc = numpy.count_nonzero(occ_idx)
        assert (nocc > 0)
        return nocc
    else:
        raise NotImplementedError

def get_nmo(mp):
    if mp._nmo is not None:
        return mp._nmo
    elif mp.frozen is None:
        return len(mp.mo_occ)
    elif isinstance(mp.frozen, (int, numpy.integer)):
        return len(mp.mo_occ) - mp.frozen
    elif hasattr(mp.frozen, '__len__'):
        return len(mp.mo_occ) - len(set(mp.frozen))
    else:
        raise NotImplementedError

def get_frozen_mask(mp):
    '''Get boolean mask for the restricted reference orbitals.

    In the returned boolean (mask) array of frozen orbital indices, the
    element is False if it corresponds to the frozen orbital.
    '''
    moidx = numpy.ones(mp.mo_occ.size, dtype=bool)
    if mp._nmo is not None:
        moidx[mp._nmo:] = False
    elif mp.frozen is None:
        pass
    elif isinstance(mp.frozen, (int, numpy.integer)):
        moidx[:mp.frozen] = False
    elif hasattr(mp.frozen, '__len__'):
        moidx[list(mp.frozen)] = False
    else:
        raise NotImplementedError
    return moidx

def get_e_hf(mp, mo_coeff=None):
    # Get HF energy, which is needed for total MP2 energy.
    if mo_coeff is None:
        mo_coeff = mp.mo_coeff
    if mo_coeff is mp._scf.mo_coeff and mp._scf.converged:
        return mp._scf.e_tot
    dm = mp._scf.make_rdm1(mo_coeff, mp.mo_occ)
    vhf = mp._scf.get_veff(mp._scf.mol, dm)
    return mp._scf.energy_tot(dm=dm, vhf=vhf)


def as_scanner(mp):
    '''Generating a scanner/solver for MP2 PES.

    The returned solver is a function. This function requires one argument
    "mol" as input and returns total MP2 energy.

    The solver will automatically use the results of last calculation as the
    initial guess of the new calculation.  All parameters assigned in the
    MP2 and the underlying SCF objects (conv_tol, max_memory etc) are
    automatically applied in the solver.

    Note scanner has side effects.  It may change many underlying objects
    (_scf, with_df, with_x2c, ...) during calculation.

    Examples::

        >>> from pyscf import gto, scf, mp
        >>> mol = gto.M(atom='H 0 0 0; F 0 0 1')
        >>> mp2_scanner = mp.MP2(scf.RHF(mol)).as_scanner()
        >>> e_tot = mp2_scanner(gto.M(atom='H 0 0 0; F 0 0 1.1'))
        >>> e_tot = mp2_scanner(gto.M(atom='H 0 0 0; F 0 0 1.5'))
    '''
    if isinstance(mp, lib.SinglePointScanner):
        return mp

    logger.info(mp, 'Set %s as a scanner', mp.__class__)
    name = mp.__class__.__name__ + MP2_Scanner.__name_mixin__
    return lib.set_class(MP2_Scanner(mp), (MP2_Scanner, mp.__class__), name)

class MP2_Scanner(lib.SinglePointScanner):
    def __init__(self, mp):
        self.__dict__.update(mp.__dict__)
        self._scf = mp._scf.as_scanner()

    def __call__(self, mol_or_geom, **kwargs):
        if isinstance(mol_or_geom, gto.MoleBase):
            mol = mol_or_geom
        else:
            mol = self.mol.set_geom_(mol_or_geom, inplace=False)

        self.reset(mol)

        mf_scanner = self._scf
        mf_scanner(mol)
        self.mo_coeff = mf_scanner.mo_coeff
        self.mo_occ = mf_scanner.mo_occ
        self.kernel(**kwargs)
        return self.e_tot


class MP2Base(lib.StreamObject):
    '''restricted MP2 with canonical HF and non-canonical HF reference

    Attributes:
        verbose : int
            Print level.  Default value equals to :class:`Mole.verbose`
        max_memory : float or int
            Allowed memory in MB.  Default value equals to :class:`Mole.max_memory`
        conv_tol : float
            For non-canonical MP2, converge threshold for MP2
            correlation energy.  Default value is 1e-7.
        conv_tol_normt : float
            For non-canonical MP2, converge threshold for
            norm(t2).  Default value is 1e-5.
        max_cycle : int
            For non-canonical MP2, max number of MP2
            iterations.  Default value is 50.
        diis_space : int
            For non-canonical MP2, DIIS space size in MP2
            iterations.  Default is 6.
        level_shift : float
            A shift on virtual orbital energies to stabilize the MP2 iterations.
        frozen : int or list
            If integer is given, the inner-most orbitals are excluded from MP2
            amplitudes.  Given the orbital indices (0-based) in a list, both
            occupied and virtual orbitals can be frozen in MP2 calculation.

            >>> mol = gto.M(atom = 'H 0 0 0; F 0 0 1.1', basis = 'ccpvdz')
            >>> mf = scf.RHF(mol).run()
            >>> # freeze 2 core orbitals
            >>> pt = mp.MP2(mf).set(frozen = 2).run()
            >>> # auto-generate the number of core orbitals to be frozen (1 in this case)
            >>> pt = mp.MP2(mf).set_frozen().run()
            >>> # freeze 2 core orbitals and 3 high lying unoccupied orbitals
            >>> pt.set(frozen = [0,1,16,17,18]).run()

    Saved results

        e_corr : float
            MP2 correlation correction
        e_corr_ss/os : float
            Same-spin and opposite-spin component of the MP2 correlation energy
        e_tot : float
            Total MP2 energy (HF + correlation)
        t2 :
            T amplitudes t2[i,j,a,b]  (i,j in occ, a,b in virt)
    '''

    # Use CCSD default settings for the moment
    max_cycle = getattr(__config__, 'cc_ccsd_CCSD_max_cycle', 50)
    conv_tol = getattr(__config__, 'cc_ccsd_CCSD_conv_tol', 1e-7)
    conv_tol_normt = getattr(__config__, 'cc_ccsd_CCSD_conv_tol_normt', 1e-5)

    _keys = {
        'max_cycle', 'conv_tol', 'conv_tol_normt', 'mol', 'max_memory',
        'frozen', 'level_shift', 'mo_coeff', 'mo_occ', 'e_hf', 'e_corr',
        'e_corr_ss', 'e_corr_os', 't2',
    }

    def __init__(self, mf, frozen=None, mo_coeff=None, mo_occ=None):

        if mo_coeff is None: mo_coeff = mf.mo_coeff
        if mo_occ is None: mo_occ = mf.mo_occ

        self.mol = mf.mol
        self._scf = mf
        self.verbose = self.mol.verbose
        self.stdout = self.mol.stdout
        self.max_memory = mf.max_memory

        self.frozen = frozen

# For iterative MP2
        self.level_shift = 0

##################################################
# don't modify the following attributes, they are not input options
        self.mo_coeff = mo_coeff
        self.mo_occ = mo_occ
        self._nocc = None
        self._nmo = None
        self.e_hf = None
        self.e_corr = None
        self.e_corr_ss = None
        self.e_corr_os = None
        self.t2 = None

    @property
    def nocc(self):
        return self.get_nocc()
    @nocc.setter
    def nocc(self, n):
        self._nocc = n

    @property
    def nmo(self):
        return self.get_nmo()
    @nmo.setter
    def nmo(self, n):
        self._nmo = n

    def reset(self, mol=None):
        if mol is not None:
            self.mol = mol
        self._scf.reset(mol)
        return self

    get_nocc = get_nocc
    get_nmo = get_nmo
    get_frozen_mask = get_frozen_mask
    get_e_hf = get_e_hf

    def set_frozen(self, method='auto', window=(-1000.0, 1000.0)):
        from pyscf import mp
        from pyscf.cc.ccsd import set_frozen
        is_gmp = isinstance(self, mp.gmp2.GMP2)
        return set_frozen(self, method=method, window=window, is_gcc=is_gmp)

    def dump_flags(self, verbose=None):
        log = logger.new_logger(self, verbose)
        log.info('')
        log.info('******** %s ********', self.__class__)
        log.info('nocc = %s, nmo = %s', self.nocc, self.nmo)
        if self.frozen is not None:
            log.info('frozen orbitals %s', self.frozen)
        log.info('max_memory %d MB (current use %d MB)',
                 self.max_memory, lib.current_memory()[0])
        return self

    @property
    def emp2(self):
        return self.e_corr

    @property
    def emp2_scs(self):
        # J. Chem. Phys. 118, 9095 (2003)
        return self.e_corr_ss*1./3. + self.e_corr_os*1.2

    @property
    def e_tot(self):
        return self.e_hf + self.e_corr

    @property
    def e_tot_scs(self):
        # J. Chem. Phys. 118, 9095 (2003)
        return self.e_hf + self.emp2_scs

    def kernel(self, mo_energy=None, mo_coeff=None, eris=None, with_t2=WITH_T2):
        '''
        Args:
            with_t2 : bool
                Whether to generate and hold t2 amplitudes in memory.
        '''
        if self.verbose >= logger.WARN:
            self.check_sanity()

        log = logger.new_logger(self)

        cput0 = cput1 = (logger.process_clock(), logger.perf_counter())

        self.dump_flags()

        self.e_hf = self.get_e_hf(mo_coeff=mo_coeff)

        cput1 = log.timer('ehf', *cput1)

        if eris is None:
            eris = self.ao2mo(mo_coeff)

        cput1 = log.timer('ao2mo', *cput1)

        if self._scf.converged:
            self.e_corr, self.t2 = self.init_amps(mo_energy, mo_coeff, eris, with_t2)
        else:
            self.converged, self.e_corr, self.t2 = _iterative_kernel(self, eris)

        cput1 = log.timer('kernel', *cput1)

        self.e_corr_ss = getattr(self.e_corr, 'e_corr_ss', 0)
        self.e_corr_os = getattr(self.e_corr, 'e_corr_os', 0)
        self.e_corr = float(self.e_corr)

        log.timer(self.__class__.__name__, *cput0)

        self._finalize()
        return self.e_corr, self.t2

    def _finalize(self):
        '''Hook for dumping results and clearing up the object.'''
        log = logger.new_logger(self)
        log.note('E(%s) = %.15g  E_corr = %.15g',
                 self.__class__.__name__, self.e_tot, self.e_corr)
        log.note('E(SCS-%s) = %.15g  E_corr = %.15g',
                 self.__class__.__name__, self.e_tot_scs, self.emp2_scs)
        log.info('E_corr(same-spin) = %.15g', self.e_corr_ss)
        log.info('E_corr(oppo-spin) = %.15g', self.e_corr_os)
        return self

    ao2mo = NotImplemented

    def make_fno(mp, thresh=1e-6, pct_occ=None, nvir_act=None, t2=None):
        raise NotImplementedError

    def make_rdm1(mp, t2=None, eris=None, ao_repr=False, with_frozen=True):
        raise NotImplementedError

    def make_rdm2(mp, t2=None, eris=None, ao_repr=False):
        raise NotImplementedError

    as_scanner = as_scanner

    def energy(self, t2, eris):
        raise NotImplementedError

    def nuc_grad_method(self):
        raise NotImplementedError

    def init_amps(self, mo_energy=None, mo_coeff=None, eris=None, with_t2=WITH_T2):
        raise NotImplementedError

    # For non-canonical MP2
    def update_amps(self, t2, eris):
        raise NotImplementedError

    def density_fit(self, auxbasis=None, with_df=None):
        raise NotImplementedError

    to_gpu = lib.to_gpu

class RMP2(MP2Base):

    make_rdm1 = make_rdm1
    make_fno = make_fno
    make_rdm2 = make_rdm2

    update_amps = update_amps

    # For non-canonical MP2
    energy = energy
    def init_amps(self, mo_energy=None, mo_coeff=None, eris=None, with_t2=WITH_T2):
        return kernel(self, mo_energy, mo_coeff, eris, with_t2)

    def ao2mo(self, mo_coeff=None):
        return _make_eris(self, mo_coeff, verbose=self.verbose)

    def density_fit(self, auxbasis=None, with_df=None):
        from pyscf.mp import dfmp2
        mymp = dfmp2.DFMP2(self._scf, self.frozen, self.mo_coeff, self.mo_occ)
        if with_df is not None:
            mymp.with_df = with_df
        if mymp.with_df.auxbasis != auxbasis:
            mymp.with_df = mymp.with_df.copy()
            mymp.with_df.auxbasis = auxbasis
        return mymp

    def nuc_grad_method(self):
        from pyscf.grad import mp2
        return mp2.Gradients(self)

MP2 = RMP2

from pyscf import scf
scf.hf.RHF.MP2 = lib.class_as_method(MP2)
scf.rohf.ROHF.MP2 = None


def _mo_energy_without_core(mp, mo_energy):
    return mo_energy[get_frozen_mask(mp)]

def _mo_without_core(mp, mo):
    return mo[:,get_frozen_mask(mp)]

def _mem_usage(nocc, nvir):
    nmo = nocc + nvir
    basic = ((nocc*nvir)**2 + nocc*nvir**2*2)*8 / 1e6
    incore = nocc*nvir*nmo**2/2*8 / 1e6 + basic
    outcore = basic
    return incore, outcore, basic

#TODO: Merge this _ChemistsERIs class with ccsd._ChemistsERIs class
class _ChemistsERIs:
    def __init__(self, mol=None):
        self.mol = mol
        self.mo_coeff = None
        self.nocc = None
        self.fock = None
        self.orbspin = None
        self.ovov = None

    def _common_init_(self, mp, mo_coeff=None):
        if mo_coeff is None:
            mo_coeff = mp.mo_coeff
        if mo_coeff is None:
            raise RuntimeError('mo_coeff, mo_energy are not initialized.\n'
                               'You may need to call mf.kernel() to generate them.')

        self.mo_coeff = _mo_without_core(mp, mo_coeff)
        self.mol = mp.mol

        if mo_coeff is mp._scf.mo_coeff and mp._scf.converged:
            # The canonical MP2 from a converged SCF result. Rebuilding fock
            # can be skipped
            self.mo_energy = _mo_energy_without_core(mp, mp._scf.mo_energy)
            self.fock = numpy.diag(self.mo_energy)
        else:
            dm = mp._scf.make_rdm1(mo_coeff, mp.mo_occ)
            vhf = mp._scf.get_veff(mp.mol, dm)
            fockao = mp._scf.get_fock(vhf=vhf, dm=dm)
            self.fock = self.mo_coeff.conj().T.dot(fockao).dot(self.mo_coeff)
            self.mo_energy = self.fock.diagonal().real
        return self

def _make_eris(mp, mo_coeff=None, ao2mofn=None, verbose=None):
    log = logger.new_logger(mp, verbose)
    time0 = (logger.process_clock(), logger.perf_counter())
    eris = _ChemistsERIs()
    eris._common_init_(mp, mo_coeff)
    mo_coeff = eris.mo_coeff

    nocc = mp.nocc
    nmo = mp.nmo
    nvir = nmo - nocc
    mem_incore, mem_outcore, mem_basic = _mem_usage(nocc, nvir)
    mem_now = lib.current_memory()[0]
    max_memory = max(0, mp.max_memory - mem_now)
    if max_memory < mem_basic:
        log.warn('Not enough memory for integral transformation. '
                 'Available mem %s MB, required mem %s MB',
                 max_memory, mem_basic)

    co = numpy.asarray(mo_coeff[:,:nocc], order='F')
    cv = numpy.asarray(mo_coeff[:,nocc:], order='F')
    if (mp.mol.incore_anyway or
        (mp._scf._eri is not None and mem_incore < max_memory)):
        log.debug('transform (ia|jb) incore')
        if callable(ao2mofn):
            eris.ovov = ao2mofn((co,cv,co,cv)).reshape(nocc*nvir,nocc*nvir)
        else:
            eris.ovov = ao2mo.general(mp._scf._eri, (co,cv,co,cv))

    elif getattr(mp._scf, 'with_df', None):
        # To handle the PBC or custom 2-electron with 3-index tensor.
        # Call dfmp2.MP2 for efficient DF-MP2 implementation.
        log.warn('DF-HF is found. (ia|jb) is computed based on the DF '
                 '3-tensor integrals.\n'
                 'You can switch to dfmp2.MP2 for better performance')
        log.debug('transform (ia|jb) with_df')
        eris.ovov = mp._scf.with_df.ao2mo((co,cv,co,cv))

    else:
        log.debug('transform (ia|jb) outcore')
        eris.feri = lib.H5TmpFile()
        #ao2mo.outcore.general(mp.mol, (co,cv,co,cv), eris.feri,
        #                      max_memory=max_memory, verbose=log)
        #eris.ovov = eris.feri['eri_mo']
        eris.ovov = _ao2mo_ovov(mp, co, cv, eris.feri, max(2000, max_memory), log)

    log.timer('Integral transformation', *time0)
    return eris

#
# the MO integral for MP2 is (ov|ov). This is the efficient integral
# (ij|kl) => (ij|ol) => (ol|ij) => (ol|oj) => (ol|ov) => (ov|ov)
#   or    => (ij|ol) => (oj|ol) => (oj|ov) => (ov|ov)
#
def _ao2mo_ovov(mp, orbo, orbv, feri, max_memory=2000, verbose=None):
    from pyscf.scf.hf import RHF
    assert isinstance(mp._scf, RHF)
    time0 = (logger.process_clock(), logger.perf_counter())
    log = logger.new_logger(mp, verbose)

    mol = mp.mol
    int2e = mol._add_suffix('int2e')
    ao2mopt = _ao2mo.AO2MOpt(mol, int2e, 'CVHFnr_schwarz_cond',
                             'CVHFsetnr_direct_scf')
    nao, nocc = orbo.shape
    nvir = orbv.shape[1]
    nbas = mol.nbas
    assert (nvir <= nao)

    ao_loc = mol.ao_loc_nr()
    dmax = max(4, min(nao/3, numpy.sqrt(max_memory*.95e6/8/(nao+nocc)**2)))
    sh_ranges = ao2mo.outcore.balance_partition(ao_loc, dmax)
    dmax = max(x[2] for x in sh_ranges)
    eribuf = numpy.empty((nao,dmax,dmax,nao))
    ftmp = lib.H5TmpFile()
    log.debug('max_memory %s MB (dmax = %s) required disk space %g MB',
              max_memory, dmax, nocc**2*(nao*(nao+dmax)/2+nvir**2)*8/1e6)

    buf_i = numpy.empty((nocc*dmax**2*nao))
    buf_li = numpy.empty((nocc**2*dmax**2))
    buf1 = numpy.empty_like(buf_li)

    fint = gto.moleintor.getints4c
    jk_blk_slices = []
    count = 0
    time1 = time0
    with lib.call_in_background(ftmp.__setitem__) as save:
        for ip, (ish0, ish1, ni) in enumerate(sh_ranges):
            for jsh0, jsh1, nj in sh_ranges[:ip+1]:
                i0, i1 = int(ao_loc[ish0]), int(ao_loc[ish1])
                j0, j1 = int(ao_loc[jsh0]), int(ao_loc[jsh1])
                jk_blk_slices.append((i0,i1,j0,j1))

                eri = fint(int2e, mol._atm, mol._bas, mol._env,
                           shls_slice=(0,nbas,ish0,ish1, jsh0,jsh1,0,nbas),
                           aosym='s1', ao_loc=ao_loc, cintopt=ao2mopt._cintopt,
                           out=eribuf)
                tmp_i = numpy.ndarray((nocc,(i1-i0)*(j1-j0)*nao), buffer=buf_i)
                tmp_li = numpy.ndarray((nocc,nocc*(i1-i0)*(j1-j0)), buffer=buf_li)
                lib.ddot(orbo.T, eri.reshape(nao,(i1-i0)*(j1-j0)*nao), c=tmp_i)
                lib.ddot(orbo.T, tmp_i.reshape(nocc*(i1-i0)*(j1-j0),nao).T, c=tmp_li)
                tmp_li = tmp_li.reshape(nocc,nocc,(i1-i0),(j1-j0))
                save(str(count), tmp_li.transpose(1,0,2,3))
                buf_li, buf1 = buf1, buf_li
                count += 1
                time1 = log.timer_debug1('partial ao2mo [%d:%d,%d:%d]' %
                                         (ish0,ish1,jsh0,jsh1), *time1)
    time1 = time0 = log.timer('mp2 ao2mo_ovov pass1', *time0)
    eri = eribuf = tmp_i = tmp_li = buf_i = buf_li = buf1 = None

    h5dat = feri.create_dataset('ovov', (nocc*nvir,nocc*nvir), 'f8',
                                chunks=(nvir,nvir))
    occblk = int(min(nocc, max(4, 250/nocc, max_memory*.9e6/8/(nao**2*nocc)/5)))
    def load(i0, eri):
        if i0 < nocc:
            i1 = min(i0+occblk, nocc)
            for k, (p0,p1,q0,q1) in enumerate(jk_blk_slices):
                eri[:i1-i0,:,p0:p1,q0:q1] = ftmp[str(k)][i0:i1]
                if p0 != q0:
                    dat = numpy.asarray(ftmp[str(k)][:,i0:i1])
                    eri[:i1-i0,:,q0:q1,p0:p1] = dat.transpose(1,0,3,2)

    def save(i0, i1, dat):
        for i in range(i0, i1):
            h5dat[i*nvir:(i+1)*nvir] = dat[i-i0].reshape(nvir,nocc*nvir)

    orbv = numpy.asarray(orbv, order='F')
    buf_prefecth = numpy.empty((occblk,nocc,nao,nao))
    buf = numpy.empty_like(buf_prefecth)
    bufw = numpy.empty((occblk*nocc,nvir**2))
    bufw1 = numpy.empty_like(bufw)
    with lib.call_in_background(load) as prefetch:
        with lib.call_in_background(save) as bsave:
            load(0, buf_prefecth)
            for i0, i1 in lib.prange(0, nocc, occblk):
                buf, buf_prefecth = buf_prefecth, buf
                prefetch(i1, buf_prefecth)
                eri = buf[:i1-i0].reshape((i1-i0)*nocc,nao,nao)

                dat = _ao2mo.nr_e2(eri, orbv, (0,nvir,0,nvir), 's1', 's1', out=bufw)
                bsave(i0, i1, dat.reshape(i1-i0,nocc,nvir,nvir).transpose(0,2,1,3))
                bufw, bufw1 = bufw1, bufw
                time1 = log.timer_debug1('pass2 ao2mo [%d:%d]' % (i0,i1), *time1)

    time0 = log.timer('mp2 ao2mo_ovov pass2', *time0)
    return h5dat

del (WITH_T2)


if __name__ == '__main__':
    from pyscf import scf
    mol = gto.Mole()
    mol.atom = [
        [8 , (0. , 0.     , 0.)],
        [1 , (0. , -0.757 , 0.587)],
        [1 , (0. , 0.757  , 0.587)]]

    mol.basis = 'cc-pvdz'
    mol.build()
    mf = scf.RHF(mol).run()

    pt = MP2(mf)
    emp2, t2 = pt.kernel()
    print(emp2 - -0.204019967288338)
    pt.max_memory = 1
    emp2, t2 = pt.kernel()
    print(emp2 - -0.204019967288338)

    pt = MP2(scf.density_fit(mf, 'weigend'))
    print(pt.kernel()[0] - -0.204254500454)

    mf = scf.RHF(mol).run(max_cycle=1)
    pt = MP2(mf)
    print(pt.kernel()[0] - -0.204479914961218)



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''Density expansion on plane waves'''


import contextlib
import numpy as np
from pyscf import lib
from pyscf import gto
from pyscf.lib import logger
from pyscf.pbc import tools
from pyscf.pbc.gto import pseudo, error_for_ke_cutoff
from pyscf.pbc import gto as pbcgto
from pyscf.pbc.gto.pseudo import pp_int
from pyscf.pbc.lib.kpts_helper import is_zero, gamma_point
from pyscf.pbc.df import ft_ao
from pyscf.pbc.df import aft_jk
from pyscf.pbc.df import aft_ao2mo
from pyscf.pbc.df.incore import Int3cBuilder
from pyscf.pbc.tools import k2gamma
from pyscf.pbc.tools import pbc as pbctools
from pyscf import __config__


KE_SCALING = getattr(__config__, 'pbc_df_aft_ke_cutoff_scaling', 0.75)
RCUT_THRESHOLD = getattr(__config__, 'pbc_scf_rsjk_rcut_threshold', 2.0)

def estimate_eta_min(cell, cutoff=None):
    '''Given rcut the boundary of repeated images of the cell, estimates the
    minimal exponent of the smooth compensated gaussian model charge, requiring
    that at boundary, density ~ 4pi rmax^2 exp(-eta/2*rmax^2) < cutoff
    '''
    from pyscf.pbc.df.gdf_builder import estimate_eta_min
    logger.warn(cell, 'Function deprecated. '
                'Call pbc.df.gdf_builder.estimate_eta_min instead.')
    return estimate_eta_min(cell, cutoff)

estimate_eta = estimate_eta_min

def estimate_eta_for_ke_cutoff(cell, ke_cutoff, precision=None):
    '''Given ke_cutoff, the upper bound of eta to produce the required
    precision in AFTDF Coulomb integrals.
    '''
    from pyscf.pbc.df.gdf_builder import estimate_eta_for_ke_cutoff
    logger.warn(cell, 'Function deprecated. '
                'Call pbc.df.gdf_builder.estimate_eta_for_ke_cutoff instead.')
    return estimate_eta_for_ke_cutoff(cell, ke_cutoff, precision)

def estimate_ke_cutoff_for_eta(cell, eta, precision=None):
    '''Given eta, the lower bound of ke_cutoff to produce the required
    precision in AFTDF Coulomb integrals.
    '''
    from pyscf.pbc.df.gdf_builder import estimate_ke_cutoff_for_eta
    logger.warn(cell, 'Function deprecated. '
                'Call pbc.df.gdf_builder.estimate_ke_cutoff_for_eta instead.')
    return estimate_ke_cutoff_for_eta(cell, eta, precision)

def estimate_omega_min(cell, cutoff=None):
    '''Given cell.rcut the boundary of repeated images of the cell, estimates
    the minimal omega for the attenuated Coulomb interactions, requiring that at
    boundary the Coulomb potential of a point charge < cutoff
    '''
    from pyscf.pbc.df.rsdf_builder import estimate_omega_min
    logger.warn(cell, 'Function deprecated. '
                'Call pbc.df.rsdf_builder.estimate_omega_min instead.')
    return estimate_omega_min(cell, cutoff)

estimate_omega = estimate_omega_min

def estimate_ke_cutoff_for_omega(cell, omega, precision=None):
    '''Energy cutoff for AFTDF to converge attenuated Coulomb in moment space
    '''
    from pyscf.pbc.df.rsdf_builder import estimate_ke_cutoff_for_omega
    logger.warn(cell, 'Function deprecated. '
                'Call pbc.df.rsdf_builder.estimate_ke_cutoff_for_omega instead.')
    return estimate_ke_cutoff_for_omega(cell, omega, precision)

def estimate_omega_for_ke_cutoff(cell, ke_cutoff, precision=None):
    '''The minimal omega in attenuated Coulombl given energy cutoff
    '''
    from pyscf.pbc.df.rsdf_builder import estimate_omega_for_ke_cutoff
    logger.warn(cell, 'Function deprecated. '
                'Call pbc.df.rsdf_builder.estimate_omega_for_ke_cutoff instead.')
    return estimate_omega_for_ke_cutoff(cell, ke_cutoff, precision)


def _get_pp_loc_part1(mydf, kpts=None, with_pseudo=True):
    kpts, is_single_kpt = _check_kpts(mydf, kpts)
    log = logger.Logger(mydf.stdout, mydf.verbose)
    t0 = t1 = (logger.process_clock(), logger.perf_counter())
    cell = mydf.cell
    mesh = np.asarray(mydf.mesh)
    nkpts = len(kpts)
    nao = cell.nao_nr()
    nao_pair = nao * (nao+1) // 2

    kpt_allow = np.zeros(3)
    if cell.dimension > 0:
        ke_guess = estimate_ke_cutoff(cell, cell.precision)
        mesh_guess = cell.cutoff_to_mesh(ke_guess)
        if np.any(mesh < mesh_guess*KE_SCALING):
            logger.warn(mydf, 'mesh %s is not enough for AFTDF.get_nuc function '
                        'to get integral accuracy %g.\nRecommended mesh is %s.',
                        mesh, cell.precision, mesh_guess)
    log.debug1('aft.get_pp_loc_part1 mesh = %s', mesh)
    Gv, Gvbase, kws = cell.get_Gv_weights(mesh)

    if with_pseudo:
        vpplocG = pp_int.get_gth_vlocG_part1(cell, Gv)
        vpplocG = -np.einsum('ij,ij->j', cell.get_SI(Gv), vpplocG)
    else:
        fakenuc = _fake_nuc(cell, with_pseudo=with_pseudo)
        aoaux = ft_ao.ft_ao(fakenuc, Gv)
        charges = cell.atom_charges()
        coulG = pbctools.get_coulG(cell, kpt_allow, mesh=mesh, Gv=Gv)
        vpplocG = np.einsum('i,xi,x->x', -charges, aoaux, coulG)

    vpplocG *= kws
    vGR = vpplocG.real
    vGI = vpplocG.imag

    vjR = np.zeros((nkpts, nao_pair))
    vjI = np.zeros((nkpts, nao_pair))
    max_memory = max(2000, mydf.max_memory-lib.current_memory()[0])
    for Gpq, p0, p1 in mydf.ft_loop(mesh, kpt_allow, kpts, aosym='s2',
                                    max_memory=max_memory, return_complex=False):
        # shape of Gpq (nkpts, nGv, nao_pair)
        for k, (GpqR, GpqI) in enumerate(zip(*Gpq)):
            # rho_ij(G) nuc(-G) / G^2
            # = [Re(rho_ij(G)) + Im(rho_ij(G))*1j] [Re(nuc(G)) - Im(nuc(G))*1j] / G^2
            vjR[k] += np.einsum('k,kx->x', vGR[p0:p1], GpqR)
            vjR[k] += np.einsum('k,kx->x', vGI[p0:p1], GpqI)
            if not is_zero(kpts[k]):
                vjI[k] += np.einsum('k,kx->x', vGR[p0:p1], GpqI)
                vjI[k] -= np.einsum('k,kx->x', vGI[p0:p1], GpqR)
        t1 = log.timer_debug1('contracting Vnuc [%s:%s]'%(p0, p1), *t1)
    log.timer_debug1('contracting Vnuc', *t0)

    vj_kpts = []
    for k, kpt in enumerate(kpts):
        if is_zero(kpt):
            vj_kpts.append(lib.unpack_tril(vjR[k]))
        else:
            vj_kpts.append(lib.unpack_tril(vjR[k]+vjI[k]*1j))
    if is_single_kpt:
        vj_kpts = vj_kpts[0]
    return np.asarray(vj_kpts)

def _check_kpts(mydf, kpts):
    '''Check if the argument kpts is a single k-point'''
    if kpts is None:
        kpts = mydf.kpts
    if kpts is None:
        kpts = np.zeros((1, 3))
        is_single_kpt = True
    else:
        kpts = np.asarray(kpts)
        is_single_kpt = kpts.ndim == 1
    kpts = kpts.reshape(-1,3)
    return kpts, is_single_kpt

def _int_nuc_vloc(mydf, nuccell, kpts, intor='int3c2e', aosym='s2', comp=1):
    '''Vnuc - Vloc'''
    raise DeprecationWarning

def get_pp(mydf, kpts=None):
    '''Get the periodic pseudopotential nuc-el AO matrix, with G=0 removed.

    Kwargs:
        mesh: custom mesh grids. By default mesh is determined by the
        function _guess_eta from module pbc.df.gdf_builder.
    '''
    t0 = (logger.process_clock(), logger.perf_counter())
    kpts, is_single_kpt = _check_kpts(mydf, kpts)
    cell = mydf.cell
    vpp = _get_pp_loc_part1(mydf, kpts, with_pseudo=True)
    t1 = logger.timer_debug1(mydf, 'get_pp_loc_part1', *t0)
    pp2builder = _IntPPBuilder(cell, kpts)
    vpp += pp2builder.get_pp_loc_part2()
    t1 = logger.timer_debug1(mydf, 'get_pp_loc_part2', *t1)
    vpp += pp_int.get_pp_nl(cell, kpts)
    t1 = logger.timer_debug1(mydf, 'get_pp_nl', *t1)
    if is_single_kpt:
        vpp = vpp[0]
    logger.timer(mydf, 'get_pp', *t0)
    return vpp


def get_nuc(mydf, kpts=None):
    '''Get the periodic nuc-el AO matrix, with G=0 removed.

    Kwargs:
        function _guess_eta from module pbc.df.gdf_builder.
    '''
    t0 = (logger.process_clock(), logger.perf_counter())
    nuc = _get_pp_loc_part1(mydf, kpts, with_pseudo=False)
    logger.timer(mydf, 'get_nuc', *t0)
    return nuc


def weighted_coulG(mydf, kpt=np.zeros(3), exx=False, mesh=None, omega=None):
    '''Weighted regular Coulomb kernel'''
    cell = mydf.cell
    if mesh is None:
        mesh = mydf.mesh
    Gv, Gvbase, kws = cell.get_Gv_weights(mesh)
    coulG = tools.get_coulG(cell, kpt, exx, mydf, mesh, Gv, omega=omega)
    coulG *= kws
    return coulG


def _fake_nuc(cell, with_pseudo=True):
    '''A fake cell with steep gaussians to mimic nuclear density
    '''
    fakenuc = cell.copy(deep=False)
    fakenuc._atm = cell._atm.copy()
    fakenuc._atm[:,gto.PTR_COORD] = np.arange(gto.PTR_ENV_START,
                                              gto.PTR_ENV_START+cell.natm*3,3)
    _bas = []
    _env = [0]*gto.PTR_ENV_START + [cell.atom_coords().ravel()]
    ptr = gto.PTR_ENV_START + cell.natm * 3
    half_sph_norm = .5/np.sqrt(np.pi)
    for ia in range(cell.natm):
        symb = cell.atom_symbol(ia)
        if with_pseudo and symb in cell._pseudo:
            pp = cell._pseudo[symb]
            rloc, nexp, cexp = pp[1:3+1]
            eta = .5 / rloc**2
        else:
            eta = 1e16
        norm = half_sph_norm/gto.gaussian_int(2, eta)
        _env.extend([eta, norm])
        _bas.append([ia, 0, 1, 1, 0, ptr, ptr+1, 0])
        ptr += 2
    fakenuc._bas = np.asarray(_bas, dtype=np.int32)
    fakenuc._env = np.asarray(np.hstack(_env), dtype=np.double)
    fakenuc.rcut = 0.1
    return fakenuc


def _estimate_ke_cutoff(alpha, l, c, precision, omega=0):
    '''Energy cutoff estimation for 4-center Coulomb repulsion integrals'''
    norm_ang = ((2*l+1)/(4*np.pi))**2
    fac = 8*np.pi**5 * c**4*norm_ang / (2*alpha)**(4*l+2) / precision
    Ecut = 20.
    if omega <= 0:
        Ecut = np.log(fac * (Ecut*.5)**(2*l-.5) + 1.) * 2*alpha
        Ecut = np.log(fac * (Ecut*.5)**(2*l-.5) + 1.) * 2*alpha
    else:
        theta = 1./(1./(2*alpha) + 1./(2*omega**2))
        Ecut = np.log(fac * (Ecut*.5)**(2*l-.5) + 1.) * theta
        Ecut = np.log(fac * (Ecut*.5)**(2*l-.5) + 1.) * theta
    return Ecut

def estimate_ke_cutoff(cell, precision=None):
    '''Energy cutoff estimation for 4-center Coulomb repulsion integrals'''
    if cell.nbas == 0:
        return 0.
    if precision is None:
        precision = cell.precision
    exps, cs = pbcgto.cell._extract_pgto_params(cell, 'max')
    ls = cell._bas[:,gto.ANG_OF]
    cs = gto.gto_norm(ls, exps)
    Ecut = _estimate_ke_cutoff(exps, ls, cs, precision)
    return Ecut.max()


class _IntPPBuilder(Int3cBuilder):
    '''3-center integral builder for pp loc part2 only
    '''
    def __init__(self, cell, kpts=np.zeros((1,3))):
        # cache ovlp_mask which are reused for different types of intor
        self._supmol = None
        self._ovlp_mask = None
        self._cell0_ovlp_mask = None
        Int3cBuilder.__init__(self, cell, None, kpts)

    def get_ovlp_mask(self, cutoff, supmol=None, cintopt=None):
        if self._ovlp_mask is None or supmol is not self._supmol:
            self._ovlp_mask, self._cell0_ovlp_mask = \
                    Int3cBuilder.get_ovlp_mask(self, cutoff, supmol, cintopt)
            self._supmol = supmol
        return self._ovlp_mask, self._cell0_ovlp_mask

    def build(self):
        pass

    def get_pp_loc_part2(self):
        log = logger.new_logger(self)
        cell = self.cell
        kpts = self.kpts
        nkpts = len(kpts)

        self.bvk_kmesh = kmesh = k2gamma.kpts_to_kmesh(cell, kpts)
        log.debug('kmesh for bvk-cell = %s', kmesh)

        self.rs_cell = rs_cell = ft_ao._RangeSeparatedCell.from_cell(
            cell, self.ke_cutoff, RCUT_THRESHOLD, verbose=log)

        intors = ('int3c2e', 'int3c1e', 'int3c1e_r2_origk',
                  'int3c1e_r4_origk', 'int3c1e_r6_origk')
        fake_cells = {}
        for cn in range(1, 5):
            fake_cell = pp_int.fake_cell_vloc(cell, cn)
            if fake_cell.nbas > 0:
                fake_cells[cn] = fake_cell

        if not fake_cells:
            if any(cell.atom_symbol(ia) in cell._pseudo for ia in range(cell.natm)):
                pass
            else:
                lib.logger.warn(cell, 'cell.pseudo was specified but its elements %s '
                                'were not found in the system.', cell._pseudo.keys())
            nao = cell.nao
            vpploc = np.zeros((nkpts, nao, nao))
            return vpploc

        rcut = self._estimate_rcut_3c1e(rs_cell, fake_cells)
        supmol = ft_ao.ExtendedMole.from_cell(rs_cell, kmesh, rcut.max(), log)
        self.supmol = supmol.strip_basis(rcut+1.)
        log.debug('sup-mol nbas = %d cGTO = %d pGTO = %d',
                  supmol.nbas, supmol.nao, supmol.npgto_nr())

        bufR = 0
        bufI = 0
        for cn, fake_cell in fake_cells.items():
            int3c = self.gen_int3c_kernel(
                intors[cn], 's2', comp=1, j_only=True, auxcell=fake_cell)
            vR, vI = int3c()
            bufR += np.einsum('...i->...', vR)
            if vI is not None:
                bufI += np.einsum('...i->...', vI)

        buf = (bufR + bufI * 1j).reshape(nkpts,-1)
        vpploc = []
        for k, kpt in enumerate(kpts):
            v = lib.unpack_tril(buf[k])
            if is_zero(kpt):  # gamma_point:
                v = v.real
            vpploc.append(v)
        return vpploc

    def _estimate_rcut_3c1e(self, cell, fake_cells):
        '''Estimate rcut for pp-loc part2 based on 3-center overlap integrals.
        '''
        precision = cell.precision
        exps = np.array([e.min() for e in cell.bas_exps()])
        if exps.size == 0:
            return np.zeros(1)

        ls = cell._bas[:,gto.ANG_OF]
        cs = gto.gto_norm(ls, exps)
        ai_idx = exps.argmin()
        ai = exps[ai_idx]
        li = cell._bas[ai_idx,gto.ANG_OF]
        ci = cs[ai_idx]

        r0 = cell.rcut  # initial guess
        rcut = []
        for lk, fake_cell in fake_cells.items():
            nuc_exps = np.hstack(fake_cell.bas_exps())
            ak_idx = nuc_exps.argmin()
            ak = nuc_exps[ak_idx]
            ck = abs(fake_cell._env[fake_cell._bas[ak_idx,gto.PTR_COEFF]])

            aij = ai + exps
            ajk = exps + ak
            aijk = aij + ak
            aijk1 = aijk**-.5
            theta = 1./(1./aij + 1./ak)
            norm_ang = ((2*li+1)*(2*ls+1))**.5/(4*np.pi)
            c1 = ci * cs * ck * norm_ang
            sfac = aij*exps/(aij*exps + ai*theta)
            rfac = ak / (aij * ajk)
            fl = 2
            fac = 2**(li+1)*np.pi**2.5 * aijk1**3 * c1 / theta * fl / precision

            r0 = (np.log(fac * r0 * (rfac*exps*r0+aijk1)**li *
                         (rfac*ai*r0+aijk1)**ls + 1.) / (sfac*theta))**.5
            r0 = (np.log(fac * r0 * (rfac*exps*r0+aijk1)**li *
                         (rfac*ai*r0+aijk1)**ls + 1.) / (sfac*theta))**.5
            rcut.append(r0)
        return np.max(rcut, axis=0)


class AFTDFMixin:

    weighted_coulG = weighted_coulG
    _int_nuc_vloc = _int_nuc_vloc

    def pw_loop(self, mesh=None, kpti_kptj=None, q=None, shls_slice=None,
                max_memory=2000, aosym='s1', blksize=None,
                intor='GTO_ft_ovlp', comp=1, bvk_kmesh=None):
        '''
        Fourier transform iterator for AO pair
        '''
        cell = self.cell
        if mesh is None:
            mesh = self.mesh
        if kpti_kptj is None:
            kpti = kptj = np.zeros(3)
        else:
            kpti, kptj = kpti_kptj
        if q is None:
            q = kptj - kpti

        ao_loc = cell.ao_loc_nr()
        Gv, Gvbase, kws = cell.get_Gv_weights(mesh)
        gxyz = lib.cartesian_prod([np.arange(len(x)) for x in Gvbase])
        ngrids = gxyz.shape[0]

        if shls_slice is None:
            shls_slice = (0, cell.nbas, 0, cell.nbas)
        if aosym == 's2':
            assert (shls_slice[2] == 0)
            i0 = ao_loc[shls_slice[0]]
            i1 = ao_loc[shls_slice[1]]
            nij = i1*(i1+1)//2 - i0*(i0+1)//2
        else:
            ni = ao_loc[shls_slice[1]] - ao_loc[shls_slice[0]]
            nj = ao_loc[shls_slice[3]] - ao_loc[shls_slice[2]]
            nij = ni*nj

        if blksize is None:
            blksize = min(max(64, int(max_memory*1e6*.75/(nij*16*comp))), 16384)
            sublk = int(blksize//4)
        else:
            sublk = blksize
        buf = np.empty(nij*blksize*comp, dtype=np.complex128)
        pqkRbuf = np.empty(nij*sublk*comp)
        pqkIbuf = np.empty(nij*sublk*comp)

        if bvk_kmesh is None:
            bvk_kmesh = k2gamma.kpts_to_kmesh(cell, [kpti, kptj])
        rcut = ft_ao.estimate_rcut(cell)
        supmol = ft_ao.ExtendedMole.from_cell(cell, bvk_kmesh, rcut.max())
        supmol = supmol.strip_basis(rcut)
        ft_kern = supmol.gen_ft_kernel(aosym, intor=intor, comp=comp)

        for p0, p1 in self.prange(0, ngrids, blksize):
            aoaoR, aoaoI = ft_kern(Gv[p0:p1], gxyz[p0:p1], Gvbase, q,
                                   kptj.reshape(1, 3), shls_slice, out=buf)
            aoaoR = aoaoR.reshape(comp,p1-p0,nij)
            aoaoI = aoaoI.reshape(comp,p1-p0,nij)

            for i0, i1 in lib.prange(0, p1-p0, sublk):
                nG = i1 - i0
                if comp == 1:
                    pqkR = np.ndarray((nij,nG), buffer=pqkRbuf)
                    pqkI = np.ndarray((nij,nG), buffer=pqkIbuf)
                    pqkR[:] = aoaoR[0,i0:i1].T
                    pqkI[:] = aoaoI[0,i0:i1].T
                else:
                    pqkR = np.ndarray((comp,nij,nG), buffer=pqkRbuf)
                    pqkI = np.ndarray((comp,nij,nG), buffer=pqkIbuf)
                    pqkR[:] = aoaoR[:,i0:i1].transpose(0,2,1)
                    pqkI[:] = aoaoI[:,i0:i1].transpose(0,2,1)
                yield (pqkR, pqkI, p0+i0, p0+i1)

    def ft_loop(self, mesh=None, q=np.zeros(3), kpts=None, shls_slice=None,
                max_memory=4000, aosym='s1', intor='GTO_ft_ovlp', comp=1,
                bvk_kmesh=None, return_complex=True):
        '''
        Fourier transform iterator for all kpti which satisfy
            2pi*N = (kpts - kpti - q)*a,  N = -1, 0, 1
        '''
        cell = self.cell
        if mesh is None:
            mesh = self.mesh
        if kpts is None:
            assert (is_zero(q))
            kpts = self.kpts
        kpts = np.asarray(kpts)
        nkpts = len(kpts)

        ao_loc = cell.ao_loc_nr()
        Gv, Gvbase, kws = cell.get_Gv_weights(mesh)
        gxyz = lib.cartesian_prod([np.arange(len(x)) for x in Gvbase])
        ngrids = gxyz.shape[0]

        if shls_slice is None:
            shls_slice = (0, cell.nbas, 0, cell.nbas)
        if aosym == 's2':
            assert (shls_slice[2] == 0)
            i0 = ao_loc[shls_slice[0]]
            i1 = ao_loc[shls_slice[1]]
            nij = i1*(i1+1)//2 - i0*(i0+1)//2
        else:
            ni = ao_loc[shls_slice[1]] - ao_loc[shls_slice[0]]
            nj = ao_loc[shls_slice[3]] - ao_loc[shls_slice[2]]
            nij = ni*nj

        if bvk_kmesh is None:
            bvk_kmesh = k2gamma.kpts_to_kmesh(cell, kpts)
        #TODO:
        # ke_cutoff = pbctools.mesh_to_cutoff(cell.lattice_vectors(), mesh)
        # rs_cell = ft_ao._RangeSeparatedCell.from_cell(cell, ke_cutoff, ft_ao.RCUT_THRESHOLD)

        rcut = ft_ao.estimate_rcut(cell)
        supmol = ft_ao.ExtendedMole.from_cell(cell, bvk_kmesh, rcut.max())
        supmol = supmol.strip_basis(rcut)
        ft_kern = supmol.gen_ft_kernel(aosym, intor=intor, comp=comp,
                                       return_complex=return_complex)

        blksize = max(16, int(max_memory*.9e6/(nij*nkpts*16*comp)))
        blksize = min(blksize, ngrids, 16384)
        buf = np.empty(nkpts*nij*blksize*comp, dtype=np.complex128)

        for p0, p1 in self.prange(0, ngrids, blksize):
            dat = ft_kern(Gv[p0:p1], gxyz[p0:p1], Gvbase, q, kpts, shls_slice, out=buf)
            yield dat, p0, p1

    @contextlib.contextmanager
    def range_coulomb(self, omega):
        '''Creates a temporary density fitting object for RSH-DF integrals.
        In this context, only LR or SR integrals for mol and auxmol are computed.
        '''
        key = '%.6f' % omega
        if key in self._rsh_df:
            rsh_df = self._rsh_df[key]
        else:
            rsh_df = self._rsh_df[key] = self.copy().reset()
            logger.info(self, 'Create RSH-DF object %s for omega=%s', rsh_df, omega)

        cell = self.cell
        auxcell = getattr(self, 'auxcell', None)

        cell_omega = cell.omega
        cell.omega = omega
        auxcell_omega = None
        if auxcell is not None:
            auxcell_omega = auxcell.omega
            auxcell.omega = omega

        assert rsh_df.cell.omega == omega
        if getattr(rsh_df, 'auxcell', None) is not None:
            assert rsh_df.auxcell.omega == omega

        try:
            yield rsh_df
        finally:
            cell.omega = cell_omega
            if auxcell_omega is not None:
                auxcell.omega = auxcell_omega


class AFTDF(lib.StreamObject, AFTDFMixin):
    '''Density expansion on plane waves
    '''

    _keys = {
        'cell', 'mesh', 'kpts', 'time_reversal_symmetry', 'blockdim',
    }

    # to mimic molecular DF object
    blockdim = getattr(__config__, 'pbc_df_df_DF_blockdim', 240)

    def __init__(self, cell, kpts=np.zeros((1,3))):
        self.cell = cell
        self.stdout = cell.stdout
        self.verbose = cell.verbose
        self.max_memory = cell.max_memory
        self.mesh = cell.mesh
        if cell.omega > 0:
            ke_cutoff = estimate_ke_cutoff_for_omega(cell, cell.omega)
            self.mesh = cell.cutoff_to_mesh(ke_cutoff)
        self.kpts = kpts
        self.time_reversal_symmetry = True

        # The following attributes are not input options.
        self._rsh_df = {}  # Range separated Coulomb DF objects

    def dump_flags(self, verbose=None):
        logger.info(self, '\n')
        logger.info(self, '******** %s ********', self.__class__)
        logger.info(self, 'mesh = %s (%d PWs)', self.mesh, np.prod(self.mesh))
        logger.info(self, 'len(kpts) = %d', len(self.kpts))
        logger.debug1(self, '    kpts = %s', self.kpts)
        return self

    def reset(self, cell=None):
        if cell is not None:
            self.cell = cell
        self._rsh_df = {}
        return self

    def check_sanity(self):
        lib.StreamObject.check_sanity(self)
        cell = self.cell
        if not cell.has_ecp():
            logger.warn(self, 'AFTDF integrals are found in all-electron '
                        'calculation.  It often causes huge error.\n'
                        'Recommended methods are DF or MDF. In SCF calculation, '
                        'they can be initialized as\n'
                        '        mf = mf.density_fit()\nor\n'
                        '        mf = mf.mix_density_fit()')

        if cell.dimension > 0:
            if cell.ke_cutoff is None:
                ke_cutoff = tools.mesh_to_cutoff(cell.lattice_vectors(), self.mesh)
                ke_cutoff = ke_cutoff[:cell.dimension].min()
            else:
                ke_cutoff = np.min(cell.ke_cutoff)
            ke_guess = estimate_ke_cutoff(cell, cell.precision)
            mesh_guess = cell.cutoff_to_mesh(ke_guess)
            if ke_cutoff < ke_guess * KE_SCALING:
                logger.warn(self, 'ke_cutoff/mesh (%g / %s) is not enough for AFTDF '
                            'to get integral accuracy %g.\nCoulomb integral error '
                            'is ~ %.2g Eh.\nRecommended ke_cutoff/mesh are %g / %s.',
                            ke_cutoff, self.mesh, cell.precision,
                            error_for_ke_cutoff(cell, ke_cutoff), ke_guess, mesh_guess)
        else:
            mesh_guess = np.copy(self.mesh)

        if cell.dimension < 3:
            err = np.exp(-0.436392335*min(self.mesh[cell.dimension:]) - 2.99944305)
            err *= cell.nelectron
            meshz = pbcgto.cell._mesh_inf_vaccum(cell)
            mesh_guess[cell.dimension:] = int(meshz)
            if err > cell.precision*10:
                logger.warn(self, 'mesh %s of AFTDF may not be enough to get '
                            'integral accuracy %g for %dD PBC system.\n'
                            'Coulomb integral error is ~ %.2g Eh.\n'
                            'Recommended mesh is %s.',
                            self.mesh, cell.precision, cell.dimension, err, mesh_guess)
            if any(x/meshz > 1.1 for x in cell.mesh[cell.dimension:]):
                meshz = pbcgto.cell._mesh_inf_vaccum(cell)
                logger.warn(self, 'setting mesh %s of AFTDF too high in non-periodic direction '
                            '(=%s) can result in an unnecessarily slow calculation.\n'
                            'For coulomb integral error of ~ %.2g Eh in %dD PBC, \n'
                            'a recommended mesh for non-periodic direction is %s.',
                            self.mesh, self.mesh[cell.dimension:], cell.precision,
                            cell.dimension, mesh_guess[cell.dimension:])
        return self

    def build(self):
        return self.check_sanity()

    get_nuc = get_nuc
    get_pp = get_pp

    # Note: Special exxdiv by default should not be used for an arbitrary
    # input density matrix. When the df object was used with the molecular
    # post-HF code, get_jk was often called with an incomplete DM (e.g. the
    # core DM in CASCI). An SCF level exxdiv treatment is inadequate for
    # post-HF methods.
    def get_jk(self, dm, hermi=1, kpts=None, kpts_band=None,
               with_j=True, with_k=True, omega=None, exxdiv=None):
        if omega is not None:  # J/K for RSH functionals
            with self.range_coulomb(omega) as rsh_df:
                return rsh_df.get_jk(dm, hermi, kpts, kpts_band, with_j, with_k,
                                     omega=None, exxdiv=exxdiv)

        kpts, is_single_kpt = _check_kpts(self, kpts)
        if is_single_kpt:
            return aft_jk.get_jk(self, dm, hermi, kpts[0], kpts_band, with_j,
                                  with_k, exxdiv)

        vj = vk = None
        if with_k:
            vk = aft_jk.get_k_kpts(self, dm, hermi, kpts, kpts_band, exxdiv)
        if with_j:
            vj = aft_jk.get_j_kpts(self, dm, hermi, kpts, kpts_band)
        return vj, vk

    get_eri = get_ao_eri = aft_ao2mo.get_eri
    ao2mo = get_mo_eri = aft_ao2mo.general
    ao2mo_7d = aft_ao2mo.ao2mo_7d
    get_ao_pairs_G = get_ao_pairs = aft_ao2mo.get_ao_pairs_G
    get_mo_pairs_G = get_mo_pairs = aft_ao2mo.get_mo_pairs_G

    def update_mf(self, mf):
        mf = mf.copy()
        mf.with_df = self
        return mf

    def prange(self, start, stop, step):
        '''This is a hook for MPI parallelization. DO NOT use it out of the
        scope of AFTDF/GDF/MDF.
        '''
        return lib.prange(start, stop, step)

################################################################################
# With this function to mimic the molecular DF.loop function, the pbc gamma
# point DF object can be used in the molecular code
    def loop(self, blksize=None):
        cell = self.cell
        if cell.dimension == 2 and cell.low_dim_ft_type != 'inf_vacuum':
            raise RuntimeError('ERIs of PBC-2D systems are not positive '
                               'definite. Current API only supports positive '
                               'definite ERIs.')

        if blksize is None:
            blksize = self.blockdim
        # coulG of 1D and 2D has negative elements.
        coulG = self.weighted_coulG()
        Lpq = None
        for pqkR, pqkI, p0, p1 in self.pw_loop(aosym='s2', blksize=blksize):
            vG = np.sqrt(coulG[p0:p1])
            pqkR *= vG
            pqkI *= vG
            Lpq = lib.transpose(pqkR, out=Lpq)
            yield Lpq
            Lpq = lib.transpose(pqkI, out=Lpq)
            yield Lpq

    def get_naoaux(self):
        mesh = np.asarray(self.mesh)
        ngrids = np.prod(mesh)
        return ngrids * 2

def _sub_df_jk_(dfobj, dm, hermi=1, kpts=None, kpts_band=None,
                with_j=True, with_k=True, omega=None, exxdiv=None):
    with dfobj.range_coulomb(omega) as rsh_df:
        return rsh_df.get_jk(dm, hermi, kpts, kpts_band, with_j, with_k,
                             omega=None, exxdiv=exxdiv)



#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np
from scipy import linalg
from pyscf import lib
from pyscf.mcscf.addons import StateAverageMCSCFSolver
from pyscf.mcscf.addons import StateAverageMixFCISolver
from pyscf.mcscf.df import _DFCASSCF
from pyscf.mcscf import mc1step
from pyscf.fci import direct_spin1
from pyscf import mcpdft
from pyscf import __config__

# API for general multi-state MC-PDFT method object
# In principle, various forms can be implemented: CMS, XMS, etc.

# API cleanup desiderata:
# 1. "sipdft", "state_interaction" -> "mspdft", "multi_state"
# 2. Canonicalize function to quickly generate mo_coeff, ci, mo_occ, mo_energy
#    for different choices of intermediate, reference, final states.
# 3. Probably "_finalize" stuff?
# 4. checkpoint stuff

MAX_CYC_DIABATIZE = getattr(__config__, 'mcpdft_mspdft_max_cyc_diabatize', 50)
CONV_TOL_DIABATIZE = getattr(__config__, 'mcpdft_mspdft_conv_tol_diabatize', 1e-8)
SING_TOL_DIABATIZE = getattr(__config__, 'mcpdft_mspdft_sing_tol_diabatize', 1e-8)
NUDGE_TOL_DIABATIZE = getattr(__config__, 'mcpdft_mspdft_nudge_tol_diabatize', 1e-3)

def make_heff_mcscf (mc, mo_coeff=None, ci=None):
    '''Build Hamiltonian matrix in basis of ci vector

    Args:
        mc : an instance of MCPDFT class

    Kwargs:
        mo_coeff : ndarray of shape (nao, nmo)
            MO coefficients
        ci : ndarray or list of len (nroots)
            CI vectors describing the model space, presumed to be in the
            optimized intermediate-state basis

    Returns:
        heff_mcscf : ndarray of shape (nroots, nroots)
            Effective MC-SCF hamiltonian matrix in the basis of the
            provided CI vectors
    '''
    if mo_coeff is None: mo_coeff = mc.mo_coeff
    if ci is None: ci = mc.ci

    h1, h0 = mc.get_h1eff (mo_coeff)
    h2 = mc.get_h2eff (mo_coeff)
    h2eff = direct_spin1.absorb_h1e (h1, h2, mc.ncas, mc.nelecas, 0.5)

    def construct_ham_slice(solver, slice, nelecas):
        ci_irrep = ci[slice]
        if hasattr(solver, "orbsym"):
            solver.orbsym = mc.fcisolver.orbsym

        hc_all_irrep = [solver.contract_2e(h2eff, c, mc.ncas, nelecas) for c in ci_irrep]
        heff_irrep = np.tensordot(ci_irrep, hc_all_irrep, axes=((1, 2), (1, 2)))
        diag_idx = np.diag_indices_from(heff_irrep)
        heff_irrep[diag_idx] += h0
        return heff_irrep

    if not isinstance(mc.fcisolver, StateAverageMixFCISolver):
        return construct_ham_slice(direct_spin1, slice(0, len(ci)), mc.nelecas)

    irrep_slices = []
    start = 0
    for solver in mc.fcisolver.fcisolvers:
        end = start+solver.nroots
        irrep_slices.append(slice(start, end))
        start = end

    return [construct_ham_slice(s, irrep, mc.fcisolver._get_nelec(s, mc.nelecas))
            for s, irrep in zip(mc.fcisolver.fcisolvers, irrep_slices)]


def si_newton (mc, ci=None, objfn=None, max_cyc=None, conv_tol=None,
        sing_tol=None, nudge_tol=None):
    '''Optimize the intermediate states describing the model space of
    an MS-PDFT calculation by maximizing the provided objective function
    using a gradient-ascent algorithm

    Args:
        mc : an instance of MSPDFT class

    Kwargs:
        ci : ndarray or list of len (nroots)
            CI vectors spanning the model space
        objfn : callable
            Takes CI vectors as a kwarg and returns the value, gradient,
            and Hessian of a chosen objective function wrt rotation
            between pairs of CI vectors
        max_cyc : integer
            Maximum number of cycles of the gradient-ascent algorithm
        conv_tol : float
            Maximum value of both gradient and step vectors at
            convergence
        sing_tol : float
            Tolerance for determining when normal coordinate belongs to
            the null space (df = d2f = 0) or when the Hessian is
            singular (df != 0, d2f = 0).
        nudge_tol : float
            Minimum step size along a normal coordinate when the surface
            is locally concave.

    Returns:
        conv : logical
            True if the optimization is converged
        ci : list of len (nroots)
            Optimized CI vectors describing intermediate states
    '''

    if ci is None: ci = mc.ci
    if objfn is None: objfn = mc.diabatizer
    if max_cyc is None: max_cyc = getattr (mc, 'max_cyc_diabatize', MAX_CYC_DIABATIZE)
    if conv_tol is None: conv_tol = getattr (mc, 'conv_tol_diabatize', CONV_TOL_DIABATIZE)
    if sing_tol is None: sing_tol = getattr (mc, 'sing_tol_diabatize', SING_TOL_DIABATIZE)
    if nudge_tol is None: nudge_tol = getattr (mc, 'nudge_tol_diabatize', NUDGE_TOL_DIABATIZE)
    ci = np.array (ci) # copy
    log = lib.logger.new_logger (mc, mc.verbose)
    nroots = mc.fcisolver.nroots
    rows,col = np.tril_indices(nroots,k=-1)
    u = np.eye (nroots)
    t = np.zeros((nroots,nroots))
    conv = False
    hdr = '{} intermediate-state'.format (mc.__class__.__name__)
    f, df, d2f, f_update = objfn (ci=ci)
    for it in range(max_cyc):
        log.info ("****iter {} ***********".format (it))
        log.info ("{} objective function value = {}".format (hdr, f))

        # Analyze Hessian
        d2f, evecs = linalg.eigh (d2f)
        evecs = np.array(evecs)
        df = np.dot (df, evecs)
        d2f_zero = np.abs (d2f) < sing_tol
        df_zero = np.abs (df) < sing_tol
        if np.any (d2f_zero & (~df_zero)):
            log.warn ("{} Hess is singular!".format (hdr))
        idx_null = d2f_zero & df_zero
        df[idx_null] = 0.0
        d2f[idx_null] = -1e-16
        pos_idx = d2f > 0
        neg_def = np.all (d2f < 0)
        log.info ("{} Hessian is negative-definite? {}".format (hdr, neg_def))

        # Analyze gradient
        grad_norm = np.linalg.norm(df)
        log.info ("{} grad norm = %f".format (hdr), grad_norm)
        log.info ("{} grad (normal modes) = {}".format (hdr, df))

        # Take step
        df[pos_idx & (np.abs (df/d2f) < nudge_tol)] = nudge_tol
        Dt = df/np.abs (d2f)
        step_norm = np.linalg.norm (Dt)
        log.info ("{} Hessian eigenvalues: {}".format (hdr, d2f))
        log.info ("{} step vector (normal modes): {}".format (hdr, Dt))
        t[:] = 0
        t[np.tril_indices(t.shape[0], k = -1)] = np.dot (Dt, evecs.T)
        t = t - t.T

        if grad_norm < conv_tol and step_norm < conv_tol and neg_def:
            conv = True
            break

        # I want the states we come from on the rows and the states we
        # are going to on the columns: |f> = |i>.Umat. However, the
        # antihermitian parameterization of a unitary operator always
        # puts it the other way around: |f> = Uop|i>, ~no matter how you
        # choose the generator indices~. So I have to transpose here.
        # Flipping the sign of t does the same thing, but don't get
        # confused: this isn't related to the choice of variables!
        u = np.dot (u, linalg.expm (t).T)
        f, df, d2f = f_update (u)

    try:
        ci = np.tensordot(u.T, ci, 1)
    except ValueError as e:
        print (u.shape, ci.shape)
        raise (e)
    if mc.verbose >= lib.logger.DEBUG:
        fmt_str = ' ' + ' '.join (['{:5.2f}',]*nroots)
        log.debug ("{} final overlap matrix:".format (hdr))
        for row in u: log.debug (fmt_str.format (*row))
    if conv:
        log.note ("{} optimization CONVERGED".format (hdr))
    else:
        log.note ("{} optimization did not converge after {} "
                   "cycles".format (hdr, it))

    return conv, list (ci)


class _MSPDFT (mcpdft.MultiStateMCPDFTSolver):
    '''MS-PDFT

    Extra attributes for MS-PDFT:

        diabatization: string
            The name describing the type of diabatization for I/O.
            Currently, only ``CMS'' is available.
        max_cyc_diabatize : integer
            Maximum cycles of the diabatization iteration. Default is 50.
        conv_tol_diabatize : float
            Convergence threshold of the diabatization algorithm. Default
            is 1e-8.
        sing_tol_diabatize : float
            Numerical tolerance for null state-rotation modes and
            singularities within the diabatization algorithm. Null modes
            (e.g., rotation between E1x and E1y states in a linear
            molecule) are ignored. Singularities (zero Hessian and
            non-zero gradient in the same mode) print a warning. Default
            is 1e-8.
        nudge_tol_diabatize : float
            Minimum step size along modes with positive curvature during
            the diabatization algorithm, so as to push away from saddle
            points and minima. Default is 1e-3.

    Saved results

        e_tot : float
            Weighted-average MS-PDFT final energy
        e_states : ndarray of shape (nroots)
            MS-PDFT final energies of the adiabatic states
        ci : list of length (nroots) of ndarrays
            CI vectors in the optimized diabatic basis. Related to the
            MC-SCF and MS-PDFT adiabat CI vectors by the expansion
            coefficients ``si_mcscf'' and ``si_pdft''. Either set of
            adiabat CI vectors can be obtained quickly via
            ``get_ci_adiabats''
        si : ndarray of shape (nroots, nroots)
            Expansion coefficients for the MS-PDFT adiabats in terms of
            the optimized diabatic states
        si_pdft : ndarray of shape (nroots, nroots)
            Synonym of si
        e_mcscf : ndarray of shape (nroots)
            Energies of the MC-SCF adiabatic states
        si_mcscf : ndarray of shape (nroots, nroots)
            Expansion coefficients for the MC-SCF adiabats in terms of
            the optimized diabatic states
        heff_mcscf : ndarray of shape (nroots, nroots)
            Molecular Hamiltonian in the diabatic basis
        hdiag_pdft : ndarray of shape (nroots)
            MC-PDFT total energies of the optimized diabatic states
    '''

    # Metaclass parent

    def __init__(self, mc, diabatizer, diabatize, diabatization):
        self.__dict__.update (mc.__dict__)
        keys = set (('diabatizer', 'diabatize', 'diabatization',
                     'heff_mcscf', 'hdiag_pdft',
                     'get_heff_offdiag', 'get_heff_pdft',
                     'si', 'si_mcscf', 'si_pdft',
                     'max_cyc_diabatize', 'conv_tol_diabatize',
                     'sing_tol_diabatize', 'nudge_tol_diabatize'))
        self._diabatizer = diabatizer
        self._diabatize = diabatize
        self._e_states = None
        self.max_cyc_diabatize = MAX_CYC_DIABATIZE
        self.conv_tol_diabatize = CONV_TOL_DIABATIZE
        self.sing_tol_diabatize = SING_TOL_DIABATIZE
        self.nudge_tol_diabatize = CONV_TOL_DIABATIZE
        self.diabatization = diabatization
        self.si_mcscf = None
        self.si_pdft = None
        self._keys = set (self.__dict__.keys ()).union (keys)

    @property
    def e_states (self):
        if self._in_mcscf_env:
            return self.fcisolver.e_states
        else:
            return self._e_states
    @e_states.setter
    def e_states (self, x):
        self._e_states = x
    # Unfixed to FCIsolver since MS-PDFT state energies are no longer
    # CI solutions

    @property
    def si (self):
        return self.si_pdft
    @si.setter
    def si (self, x):
        self.si_pdft = x

    def get_heff_offdiag (self):
        '''The off-diagonal elements of the effective Hamiltonian matrix

        = heff_mcscf - np.diag (heff_mcscf.diagonal ())
        = ( 0     H_10^* ... )
          ( H_10  0      ... )
          ( ...   ...    ... )

        Returns:
            heff_offdiag : ndarray of shape (nroots, nroots)
                Contains molecular Hamiltonian elements on the off-diagonals
                and zero on the diagonals
        '''
        idx = np.diag_indices_from (self.heff_mcscf)
        heff_offdiag = self.heff_mcscf.copy ()
        heff_offdiag[idx] = 0.0
        return heff_offdiag

    def get_heff_pdft (self):
        '''The MS-PDFT effective Hamiltonian matrix

        = get_heff_offdiag () + np.diag (hdiag_pdft)
        = ( EPDFT_0  H_10^*   ... )
          ( H_10     EPDFT_1  ... )
          ( ...      ...      ... )

        Returns:
            heff_pdft : ndarray of shape (nroots, nroots)
                Contains molecular Hamiltonian elements on the off-diagonals
                and PDFT energies on the diagonals
        '''
        idx = np.diag_indices_from (self.heff_mcscf)
        heff_pdft = self.heff_mcscf.copy ()
        heff_pdft[idx] = self.hdiag_pdft
        return heff_pdft

    def get_ci_adiabats (self, ci=None, uci='MSPDFT'):
        ''' Get the CI vectors in an alternate basis (usually one of the
            two adiabatic bases: MCSCF or MSPDFT)

            Kwargs:
                ci : list of length nroots
                    Diabatic ci vectors; defaults to self.ci
                uci : 'MSPDFT', 'MCSCF', or square array of length nroots
                    (String indicating) unitary matrix for transforming
                    ci vectors

            Returns:
                ci : list of length nroots
                    CI vectors for adiabats
        '''
        si_dict = {'MCSCF': self.si_mcscf,
                   'MSPDFT': self.si_pdft}
        if isinstance (uci, (str,np.bytes_)):
            if uci.upper () in si_dict:
                uci = si_dict[uci.upper ()]
            else:
                raise RuntimeError ("valid uci : 'MCSCF', 'MSPDFT', or ndarray")
        if ci is None: ci = self.ci
        return list (np.tensordot (uci.T, np.asarray (ci), axes=1))
    get_ci_basis=get_ci_adiabats

    def kernel (self, mo_coeff=None, ci0=None, otxc=None, grids_level=None,
                grids_attr=None, **kwargs):
        self.otfnal.reset (mol=self.mol) # scanner mode safety
        if ci0 is None and isinstance (getattr (self, 'ci', None), list):
            ci0 = [c.copy () for c in self.ci]
        self.optimize_mcscf_(mo_coeff=mo_coeff, ci0=ci0)
        diab_conv, self.ci = self.diabatize (ci=self.ci, ci0=ci0, **kwargs)
        self.converged = self.converged and diab_conv
        self.heff_mcscf = self.make_heff_mcscf ()
        e_mcscf, self.si_mcscf = self._eig_si (self.heff_mcscf)
        if abs (linalg.norm (self.e_mcscf-e_mcscf)) > 1e-9:
            raise RuntimeError (("Sanity fault: e_mcscf ({}) != "
                                "self.e_mcscf ({})").format (e_mcscf,
                                self.e_mcscf))
        self.hdiag_pdft = self.compute_pdft_energy_(
            otxc=otxc, grids_level=grids_level, grids_attr=grids_attr)[-1]
        self.e_states, self.si_pdft = self._eig_si (self.get_heff_pdft ())
        self.e_tot = np.dot (self.e_states, self.weights)
        self._log_diabats ()
        self._log_adiabats ()
        return (self.e_tot, self.e_ot, self.e_mcscf, self.e_cas, self.ci,
            self.mo_coeff, self.mo_energy)

    def optimize_mcscf_(self, mo_coeff=None, ci0=None, **kwargs):
        # Initialize in an adiabatic basis
        if ci0 is not None:
            if mo_coeff is None: mo_coeff = self.mo_coeff
            heff_mcscf = self.make_heff_mcscf (mo_coeff, ci0)
            e, self.si_mcscf = self._eig_si (heff_mcscf)
            ci1 = self.get_ci_adiabats (ci=ci0, uci='MCSCF')
        else:
            ci1 = None
        return super().optimize_mcscf_(mo_coeff=mo_coeff, ci0=ci1, **kwargs)

    # All of the below probably need to be wrapped over solvers in
    # multi-state-mix metaclass

    def diabatize (self, ci=None, ci0=None, **kwargs):
        '''Optimize the ``intermediate'' diabatic states of an MS-PDFT
        calculation in the space defined by the MC-SCF ``reference''
        adiabatic states.

        Kwargs
            ci : list of ndarrays of length nroots
                CI vectors defining the model space, usually from a
                or CASSCF kernel call
            ci0 : list of ndarrays of length nroots
                Initial guess for optimized diabatic CI vectors,
                possibly spanning a slightly different space, usually at
                a different geometry (i.e., during a geometry
                optimization or dynamics trajectory)

        Returns
            conv : logical
                Reports whether the diabatization algorithm converged
                successfully
            ci : list of ndarrays of length = nroots
                CI vectors of the optimized diabatic states
        '''
        if ci is None: ci = self.ci
        if ci0 is not None:
            ovlp = np.tensordot (np.asarray (ci).conj (), np.asarray (ci0),
                                 axes=((1,2),(1,2)))
            u, svals, vh = linalg.svd (ovlp)
            ci = self.get_ci_basis (ci=ci, uci=np.dot (u,vh))
        return self._diabatize (self, ci=ci, **kwargs)

    def diabatizer (self, mo_coeff=None, ci=None):
        '''Computes the value, gradient vector, and Hessian matrix with
        respect to pairwise state rotations of the objective function
        maximized by the optimized diabatic (``intermediate'') states.

        Kwargs
            mo_coeff : ndarray of shape (nao,nmo)
                MO coefficients
            ci : list of ndarrays of length nroots
                CI vectors

        Returns:
            f : float
                Objective function value for states
            df : ndarray of shape (npairs = nroots*(nroots-1)/2)
                Gradient vector of objective function with respect to
                pairwise rotations between states
            d2f : ndarray of shape (npairs,npairs)
                Hessian matrix of objective function with respect to
                pairwise rotations between states
            f_update : callable
                Takes a unitary matrix and returns f, df, and d2f as
                above. Some kinds of MS-PDFT can be sped up using
                intermediates. If so, the _diabatizer function can
                provide f_update. Otherwise, it defaults to running
                _diabatizer itself on a rotated set of CI vectors.
        '''
        if mo_coeff is None: mo_coeff = self.mo_coeff
        if ci is None: ci = self.ci
        rets = self._diabatizer (self, mo_coeff=mo_coeff, ci=ci)
        f, df, d2f = rets[:3]
        if len (rets) > 3 and callable (rets[3]):
            f_update = rets[3]
        else:
            def f_update (u=1):
                ci1 = self.get_ci_basis(ci=ci, uci=u)
                f1, df1, d2f1 = self._diabatizer (
                    self, mo_coeff=mo_coeff, ci=ci1)
                return f1, df1, d2f1
        return f, df, d2f, f_update

    def _eig_si (self, heff):
        return linalg.eigh (heff)

    make_heff_mcscf = make_heff_mcscf

    def _log_diabats (self):
        # Information about the intermediate states
        hdiag_mcscf = self.heff_mcscf.diagonal ()
        hdiag_pdft = self.hdiag_pdft
        nroots = len (hdiag_pdft)
        log = lib.logger.new_logger (self, self.verbose)
        f, df, d2f = self.diabatizer ()[:3]
        hdr = '{} diabatic (intermediate)'.format (self.__class__.__name__)
        log.note ('%s objective function  value = %.15g |grad| = %.7g', hdr, f, linalg.norm (df))
        log.note ('%s average energy  EPDFT = %.15g  EMCSCF = %.15g', hdr,
                  np.dot (self.weights, hdiag_pdft), np.dot (self.weights, hdiag_mcscf))
        log.note ('%s states:', hdr)
        if getattr (self.fcisolver, 'spin_square', None):
            ss = self.fcisolver.states_spin_square (self.ci, self.ncas,
                                                    self.nelecas)[0]
            for i in range (nroots):
                log.note ('  State %d  EPDFT = %.15g  EMCSCF = %.15g'
                    '  S^2 = %.7f', i, hdiag_pdft[i],
                    hdiag_mcscf[i], ss[i])
        else:
            for i in range (nroots):
                log.note ('  State %d  EPDFT = %.15g  EMCSCF = '
                    '%.15g', i, hdiag_pdft[i], hdiag_mcscf[i])
        log.info ('MS-PDFT effective Hamiltonian matrix in diabatic basis:')
        fmt_str = ' '.join (['{:9.5f}',]*nroots)
        for row in self.get_heff_pdft(): log.info (fmt_str.format (*row))
        log.info ('Diabatic states (columns) in terms of reference states '
            '(rows):')
        for row in self.si_mcscf.T: log.info (fmt_str.format (*row))

    def _log_adiabats (self):
        # Information about the final states
        log = lib.logger.new_logger (self, self.verbose)
        nroots = len (self.e_states)
        log.note ('%s adiabatic (final) states:', self.__class__.__name__)
        if getattr (self.fcisolver, 'spin_square', None):
            ci = np.tensordot (self.si.T, np.asarray (self.ci), axes=1)
            ss = self.fcisolver.states_spin_square (ci, self.ncas,
                                                    self.nelecas)[0]
            for i in range (nroots):
                log.note ('  State %d weight %g  EMSPDFT = %.15g  S^2 = %.7f',
                          i, self.weights[i], self.e_states[i], ss[i])
        else:
            for i in range (nroots):
                log.note ('  State %d weight %g  EMSPDFT = %.15g', i,
                          self.weights[i], self.e_states[i])

    def nuc_grad_method (self):
        if not isinstance (self, mc1step.CASSCF):
            raise NotImplementedError ("CASCI-based PDFT nuclear gradients")
        elif getattr (self, 'frozen', None) is not None:
            raise NotImplementedError ("PDFT nuclear gradients with frozen orbitals")
        elif isinstance (self, _DFCASSCF):
            from pyscf.df.grad.mspdft import Gradients
        else:
            from pyscf.grad.mspdft import Gradients
        return Gradients (self)

    def nac_method(self):
        if not isinstance(self, mc1step.CASSCF):
            raise NotImplementedError("CASCI-based PDFT NACs")
        elif getattr(self, 'frozen', None) is not None:
            raise NotImplementedError("PDFT NACs with frozen orbitals")
        elif isinstance(self, _DFCASSCF):
            raise NotImplementedError("PDFT NACs with density fitting")
        else:
            from pyscf.nac.mspdft import NonAdiabaticCouplings

        return NonAdiabaticCouplings(self)

    def dip_moment (self, unit='Debye', origin='Coord_Center', state=None):
        if not isinstance (self, mc1step.CASSCF):
            raise NotImplementedError ("CASCI-based PDFT dipole moments")
        elif getattr (self, 'frozen', None) is not None:
            raise NotImplementedError ("PDFT dipole moments with frozen orbitals")
        elif isinstance (self, _DFCASSCF):
            raise NotImplementedError ("PDFT dipole moments with density-fitting ERIs")
        # Monkeypatch for double prop folders
        # TODO: more elegant solution
        import os
        mypath = os.path.dirname (os.path.dirname (os.path.abspath (__file__)))
        myproppath = os.path.join (mypath, 'prop')
        # suppress irrelevant warnings when 'properties' ext mod installed
        import warnings
        with warnings.catch_warnings ():
            warnings.filterwarnings (
                "ignore", message="Module.*is under testing")
            from pyscf import prop
        prop.__path__.append (myproppath)
        prop.__path__=list(set(prop.__path__))
        from pyscf.prop.dip_moment.mspdft import ElectricDipole
        if not lib.isinteger (state):
            raise RuntimeError ('Permanent dipole requires a single state')
        dip_obj =  ElectricDipole(self)
        mol_dipole = dip_obj.kernel (state=state, unit=unit, origin=origin)
        return mol_dipole

    def trans_moment (self, unit='Debye', origin='Coord_Center', state=None):
        if not isinstance (self, mc1step.CASSCF):
            raise NotImplementedError ("CASCI-based PDFT dipole moments")
        elif getattr (self, 'frozen', None) is not None:
            raise NotImplementedError ("PDFT dipole moments with frozen orbitals")
        elif isinstance (self, _DFCASSCF):
            raise NotImplementedError ("PDFT dipole moments with density-fitting ERIs")
        # Monkeypatch for double prop folders
        # TODO: more elegant solution
        import os
        mypath = os.path.dirname (os.path.dirname (os.path.abspath (__file__)))
        myproppath = os.path.join (mypath, 'prop')
        # suppress irrelevant warnings when 'properties' ext mod installed
        import warnings
        with warnings.catch_warnings ():
            warnings.filterwarnings (
                "ignore", message="Module.*is under testing")
            from pyscf import prop
        prop.__path__.append (myproppath)
        prop.__path__=list(set(prop.__path__))
        from pyscf.prop.trans_dip_moment.mspdft import TransitionDipole
        if not hasattr(state, '__len__') or len(state) !=2:
            raise RuntimeError ('Transition dipole requires two states')
        tran_dip_obj = TransitionDipole(self)
        mol_trans_dipole = tran_dip_obj.kernel (state=state, unit=unit, origin=origin)
        return mol_trans_dipole

def get_diabfns (obj):
    '''Interpret the name of the MS-PDFT method as a pair of functions
    which optimize the intermediate states and calculate the power
    series in the corresponding objective function to second order.

    Args:
        obj : string
            Specify particular MS-PDFT method. Currently, only "CMS" is
            supported. Not case-sensitive.

    Returns:
        diabatizer : callable
            Takes model-space CI vectors in a trial intermediate-state
            basis and returns the value and first and second derivatives
            of the objective function specified by obj
        diabatize : callable
            Takes model-space CI vectors and returns CI vectors in the
            optimized intermediate-state basis
    '''

    if obj.upper () == 'CMS':
        from pyscf.mcpdft.cmspdft import e_coul as diabatizer
        diabatize = si_newton

    elif obj.upper() == "XMS":
        from pyscf.mcpdft.xmspdft import safock_energy as diabatizer
        from pyscf.mcpdft.xmspdft import solve_safock as diabatize

    else:
        raise RuntimeError ('MS-PDFT type not supported')

    return diabatizer, diabatize

def multi_state (mc, weights=(0.5,0.5), diabatization='CMS', **kwargs):
    ''' Build multi-state MC-PDFT method object

    Args:
        mc : instance of class _PDFT

    Kwargs:
        weights : sequence of floats
        diabatization : objective-function type
            Currently supports only 'cms'

    Returns:
        si : instance of class _MSPDFT
    '''

    if isinstance (mc, mcpdft.MultiStateMCPDFTSolver):
        raise RuntimeError ('already a multi-state PDFT solver')
    if isinstance (mc.fcisolver, StateAverageMixFCISolver):
        raise RuntimeError ('state-average mix type')
    if not isinstance (mc, StateAverageMCSCFSolver):
        base_name = mc.__class__.__name__
        mc = mc.state_average (weights=weights, **kwargs)
    else:
        base_name = mc.__class__.__bases__[0].__name__
    mcbase_class = mc.__class__
    diabatizer, diabatize = get_diabfns (diabatization)

    class MSPDFT (_MSPDFT, mcbase_class):
        pass
    MSPDFT.__name__ = diabatization.upper () + base_name
    return MSPDFT (mc, diabatizer, diabatize, diabatization)


if __name__ == '__main__':
    # This ^ is a convenient way to debug code that you are working on. The
    # code in this block will only execute if you run this python script as the
    # input directly: "python mspdft.py".

    from pyscf import scf, gto
    from pyscf.tools import molden # My version is better for MC-SCF
    from pyscf.fci import csf_solver
    xyz = '''O  0.00000000   0.08111156   0.00000000
             H  0.78620605   0.66349738   0.00000000
             H -0.78620605   0.66349738   0.00000000'''
    mol = gto.M (atom=xyz, basis='sto-3g', symmetry=False, output='mspdft.log',
        verbose=lib.logger.DEBUG)
    mf = scf.RHF (mol).run ()
    mc = mcpdft.CASSCF (mf, 'tPBE', 4, 4).set (fcisolver = csf_solver (mol, 1))
    mc = mc.multi_state ([1.0/3,]*3, 'cms').run ()


